# copyright ############################### #
# This file is part of the Xtrack Package.  #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

import pathlib

import numpy as np
from scipy.constants import c as clight
from scipy.constants import e as qe
from scipy.constants import epsilon_0, hbar

import xobjects as xo
import xpart as xp
import xtrack as xt
from xobjects.test_helpers import for_all_test_contexts, fix_random_seed

import pytest

test_data_folder = pathlib.Path(
        __file__).parent.joinpath('../test_data').absolute()


@for_all_test_contexts
@pytest.mark.parametrize('thick', [False, True])
@fix_random_seed(645284)
def test_radiation(test_context, thick):

    print(f"Test {test_context.__class__}")

    ctx2np = test_context.nparray_from_context_array

    L_bend = 1.
    B_T = 2

    delta = 0
    particles_ave = xp.Particles(
            _context=test_context,
            p0c=5e9 / (1 + delta), # 5 GeV
            x=np.zeros(100000),
            px=1e-4,
            py=-1e-4,
            delta=delta,
            mass0=xp.ELECTRON_MASS_EV)
    particles_ave_0 = particles_ave.copy()
    gamma = ctx2np((particles_ave.energy/particles_ave.mass0))[0]
    gamma0 = ctx2np(particles_ave.gamma0)[0]
    particles_rnd = particles_ave.copy()

    P0_J = ctx2np(particles_ave.p0c[0]) / clight * qe
    h_bend = B_T * qe / P0_J
    theta_bend = h_bend * L_bend

    if thick:
        dipole_ave = xt.Bend(length=L_bend, angle=theta_bend, k0_from_h=True,
                             radiation_flag=1, _context=test_context)
        dipole_rnd = xt.Bend(length=L_bend, angle=theta_bend, k0_from_h=True,
                             radiation_flag=2, _context=test_context)
    else:
        dipole_ave = xt.Multipole(knl=[theta_bend], length=L_bend, hxl=theta_bend,
                                radiation_flag=1, _context=test_context)
        dipole_rnd = xt.Multipole(knl=[theta_bend], length=L_bend, hxl=theta_bend,
                                radiation_flag=2, _context=test_context)

    dct_ave_before = particles_ave.to_dict()
    dct_rng_before = particles_rnd.to_dict()

    particles_ave._init_random_number_generator()
    particles_rnd._init_random_number_generator()

    dipole_ave.track(particles_ave)
    dipole_rnd.track(particles_rnd)

    dct_ave = particles_ave.to_dict()
    dct_rng = particles_rnd.to_dict()

    xo.assert_allclose(dct_ave['delta'], np.mean(dct_rng['delta']),
                    atol=0, rtol=5e-3)

    rho_0 = L_bend/theta_bend
    mass0_kg = (dct_ave['mass0']*qe/clight**2)
    r0 = qe**2/(4*np.pi*epsilon_0*mass0_kg*clight**2)
    Ps = (2 * r0 * clight * mass0_kg * clight**2 * gamma0**2 * gamma**2)/(3*rho_0**2) # W

    Delta_E_eV = -Ps*(L_bend/clight) / qe
    Delta_E_trk = (dct_ave['ptau']-dct_ave_before['ptau'])*dct_ave['p0c']

    xo.assert_allclose(Delta_E_eV, Delta_E_trk, atol=0, rtol=4e-5)

    # Check photons
    line=xt.Line(elements=[
                xt.Drift(length=1.0),
                xt.Multipole(knl=[theta_bend], length=L_bend, hxl=theta_bend),
                xt.Drift(length=1.0),
                xt.Multipole(knl=[theta_bend], length=L_bend, hxl=theta_bend)
                ])
    line.build_tracker(_context=test_context)
    line.configure_radiation(model='quantum')

    sum_photon_energy = 0
    sum_photon_energy_sq = 0
    tot_n_recorded = 0
    for iter in range(10):
        record_capacity = int(10e6)
        record = line.start_internal_logging_for_elements_of_type(xt.Multipole,
                                                                    capacity=record_capacity)
        particles_test = particles_ave_0.copy()
        particles_test_before = particles_test.copy()
        line.track(particles_test)

        particles_test.move(xo.context_default)
        particles_test_before.move(xo.context_default)
        particles_ave.move(xo.context_default)
        record.move(xo.context_default)

        Delta_E_test = (particles_test.ptau - particles_test_before.ptau
                                                            )*particles_test.p0c
        n_recorded = record._index.num_recorded
        assert n_recorded < record_capacity
        xo.assert_allclose(-np.sum(Delta_E_test),
                        np.sum(record.photon_energy[:n_recorded]),
                        atol=0, rtol=1e-6)

        sum_photon_energy += np.sum(record.photon_energy[:n_recorded])
        sum_photon_energy_sq += np.sum(record.photon_energy[:n_recorded]**2)
        tot_n_recorded += n_recorded

    p0_J = particles_ave.p0c[0] / clight * qe
    B_T = p0_J / qe / rho_0
    mass_0_kg = particles_ave.mass0 * qe / clight**2
    E_crit_J = 3 * qe * hbar * gamma**2 * B_T / (2 * mass_0_kg)

    E_ave_J = 8 * np.sqrt(3) / 45 * E_crit_J
    E_ave_eV = E_ave_J / qe

    E_sq_ave_J = 11 / 27 * E_crit_J**2
    E_sq_ave_eV = E_sq_ave_J / qe**2

    mean_photon_energy = sum_photon_energy / tot_n_recorded
    mean_photon_energy_sq = sum_photon_energy_sq / tot_n_recorded
    std_photon_energy = np.sqrt(mean_photon_energy_sq - mean_photon_energy**2)

    xo.assert_allclose(mean_photon_energy, E_ave_eV, rtol=1e-2, atol=0)
    xo.assert_allclose(std_photon_energy,
                      np.sqrt(E_sq_ave_eV - E_ave_eV**2), rtol=2e-3, atol=0)


@for_all_test_contexts
@pytest.mark.parametrize('thick', [False, True])
@fix_random_seed(8438475)
def test_ring_with_radiation(test_context, thick):

    from cpymad.madx import Madx

    # Import thick sequence
    mad = Madx(stdout=False)

    # CLIC-DR
    mad.call(str(test_data_folder.joinpath('clic_dr/sequence.madx')))
    mad.use('ring')

    # Twiss
    twthick = mad.twiss().dframe()

    # Emit
    mad.sequence.ring.beam.radiate = True
    mad.emit()
    mad_emit_table = mad.table.emit.dframe()
    mad_emit_summ = mad.table.emitsumm.dframe()

    # Makethin
    if not thick:
        mad.input(f'''
        select, flag=MAKETHIN, SLICE=4, thick=false;
        select, flag=MAKETHIN, pattern=wig, slice=1;
        select, flag=makethin, class=rfcavity, slice=1;
        MAKETHIN, SEQUENCE=ring, MAKEDIPEDGE=true;
        use, sequence=RING;
        ''')
        mad.use('ring')
    mad.twiss()

    # Build xtrack line
    line = xt.Line.from_madx_sequence(mad.sequence['RING'])
    line.particle_ref = xp.Particles(
            mass0=xp.ELECTRON_MASS_EV,
            q0=-1,
            gamma0=mad.sequence.ring.beam.gamma)

    # Build tracker
    line.matrix_stability_tol = 1e-2
    line.build_tracker()

    line.configure_radiation(model='mean')

    # Twiss
    tw = line.twiss(eneloss_and_damping=True)

    # Checks
    met = mad_emit_table

    xo.assert_allclose(tw['eneloss_turn'], mad_emit_summ.u0.iloc[0]*1e9,
                    rtol=3e-3, atol=0)
    xo.assert_allclose(tw['damping_constants_s'][0],
        met[met.loc[:, 'parameter']=='damping_constant']['mode1'].iloc[0],
        rtol=5e-3, atol=0
        )
    xo.assert_allclose(tw['damping_constants_s'][1],
        met[met.loc[:, 'parameter']=='damping_constant']['mode2'].iloc[0],
        rtol=1e-3, atol=0
        )
    xo.assert_allclose(tw['damping_constants_s'][2],
        met[met.loc[:, 'parameter']=='damping_constant']['mode3'].iloc[0],
        rtol=5e-3, atol=0
        )

    xo.assert_allclose(tw['partition_numbers'][0],
        met[met.loc[:, 'parameter']=='damping_partion']['mode1'].iloc[0],
        rtol=5e-3, atol=0
        )
    xo.assert_allclose(tw['partition_numbers'][1],
        met[met.loc[:, 'parameter']=='damping_partion']['mode2'].iloc[0],
        rtol=1e-3, atol=0
        )
    xo.assert_allclose(tw['partition_numbers'][2],
        met[met.loc[:, 'parameter']=='damping_partion']['mode3'].iloc[0],
        rtol=5e-3, atol=0
        )

    line.configure_radiation(model='mean')
    part_co = line.find_closed_orbit()

    par_for_emit = line.build_particles(
                                x_norm=50*[0],
                                zeta=part_co.zeta[0], delta=part_co.delta[0],
                                _context=test_context
                                )
    line.discard_tracker()
    line.build_tracker(test_context)
    line.configure_radiation(model='quantum')
    num_turns=1500
    line.track(par_for_emit, num_turns=num_turns, turn_by_turn_monitor=True)
    mon = line.record_last_track

    xo.assert_allclose(np.std(mon.zeta[:, 750:]),
        np.sqrt(met[met.loc[:, 'parameter']=='emittance']['mode3'].iloc[0] * np.abs(tw['bets0'])),
        rtol=0.2, atol=0
        )

    xo.assert_allclose(np.std(mon.x[:, 750:]),
        np.sqrt(met[met.loc[:, 'parameter']=='emittance']['mode1'].iloc[0] * tw['betx'][0]),
        rtol=0.2, atol=0
        )

    assert np.all(mon.y[:] < 1e-15)

    # Test particles generation (with electrons)
    bunch_intensity = 1e11
    sigma_z = 5e-3
    n_part = int(5e5)
    nemitt_x = 0.5e-6
    nemitt_y = 0.5e-6

    line.discard_tracker()
    line.build_tracker(_context=xo.context_default)
    line.configure_radiation(model='mean')
    pgen = xp.generate_matched_gaussian_bunch(line=line,
            num_particles=n_part, total_intensity_particles=bunch_intensity,
            nemitt_x=nemitt_x, nemitt_y=nemitt_y, sigma_z=sigma_z,
            _context=test_context)

    assert pgen._buffer.context is test_context
    pgen.move(_context=xo.ContextCpu())

    xo.assert_allclose(np.std(pgen.x),
                    np.sqrt(tw['dx'][0]**2*np.std(pgen.delta)**2
                            + tw['betx'][0]*nemitt_x/mad.sequence.ring.beam.gamma),
                    atol=0, rtol=1e-2)

    xo.assert_allclose(np.std(pgen.y),
                    np.sqrt(tw['dy'][0]**2*np.std(pgen.delta)**2
                            + tw['bety'][0]*nemitt_y/mad.sequence.ring.beam.gamma),
                    atol=0, rtol=1e-2)

    xo.assert_allclose(np.std(pgen.zeta), sigma_z, atol=0, rtol=5e-3)



# copyright ############################### #
# This file is part of the Xtrack Package.  #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

import json

import numpy as np

import xtrack as xt
import xobjects as xo

env = xt.Environment()

# Build a beam line
line = env.new_line(components=[
    env.new('m1', xt.Multipole, knl=np.array([1.,2.,3.])),
    env.new('d1', xt.Drift, length=2.),
    env.new('c1', xt.Cavity, frequency=400e9, voltage=1e6),
    env.new('m2', xt.Multipole, knl=np.array([1.,2.,3.])),
    env.new('c2', xt.Drift, length=2.),
    ],
)

# Save to json
line.to_json('line.json')

# Load from json
line_2 = xt.Line.from_json('line.json')

# Alternatively the to_dict method can be used, which is more flexible for
# example to save additional information in the json file

#Save
dct = line.to_dict()
dct['my_additional_info'] = 'Important information'
xt.json.dump(dct, 'line.json')

# Load
loaded_dct = xt.json.load('line.json')
line_2 = xt.Line.from_dict(loaded_dct)
# loaded_dct['my_additional_info'] contains "Important information"


from cpymad.madx import Madx
import numpy as np

import xtrack as xt

# machine = 'psb'
machine = 'ps'
# machine = 'elena'
machine = 'leir'
# machine = 'pimms'

test_data_folder = '../../test_data/'
mad = Madx()

if machine == 'leir':
    mad.call(test_data_folder + 'leir/leir_inj_nominal.beam')
    mad.input('BRHO = BEAM->PC * 3.3356 * (208./54.)')
    mad.call(test_data_folder + 'leir/leir.seq')
    mad.call(test_data_folder + 'leir/leir_inj_nominal.str')
    mad.use('leir')
    seq = mad.sequence.leir
    def_expr = False
elif machine == 'elena':
    mad.call(test_data_folder + 'elena/elena.seq')
    mad.call(test_data_folder + 'elena/highenergy.str')
    mad.call(test_data_folder + 'elena/highenergy.beam')
    mad.use('elena')
    seq = mad.sequence.elena
    def_expr = False
elif machine == 'psb':
    mad.call(test_data_folder + 'psb_chicane/psb.seq')
    mad.call(test_data_folder + 'psb_chicane/psb_fb_lhc.str')
    mad.input('beam, sequence=psb1, particle=proton, energy=1.098')
    mad.use('psb1')
    seq = mad.sequence.psb1
    def_expr = False
elif machine == 'ps':
    mad.call(test_data_folder + 'ps_sftpro/ps.seq')
    mad.call(test_data_folder + 'ps_sftpro/ps_hs_sftpro.str')
    mad.input('beam, particle=proton, pc = 14.0; BRHO = BEAM->PC * 3.3356;')
    mad.use('ps')
    seq = mad.sequence.ps
    def_expr = False
elif machine == 'pimms':
    mad.call(test_data_folder + 'pimms/PIMM.seq')
    mad.call(test_data_folder + 'pimms/betatron.str')
    mad.beam(particle='proton', gamma=1.05328945) # 50 MeV
    mad.use('pimms')
    seq = mad.sequence.pimms
    def_expr = True
else:
    raise ValueError(f'Unknown machine `{machine}`')


line = xt.Line.from_madx_sequence(seq, deferred_expressions=def_expr)
line.particle_ref = xt.Particles(gamma0=seq.beam.gamma,
                                 mass0=seq.beam.mass * 1e9,
                                 q0=seq.beam.charge)

# To see functions inside thick elements, we need to slice them
# line.slice_thick_elements(
#     slicing_strategies=[
#         xt.Strategy(slicing=None), # don't touch other elements
#         xt.Strategy(slicing=xt.Uniform(10, mode='thick'), element_type=xt.Bend),
#         xt.Strategy(slicing=xt.Uniform(5, mode='thick'), element_type=xt.Quadrupole),
#         xt.Strategy(slicing=xt.Uniform(5, mode='thick'), element_type=xt.CombinedFunctionMagnet),
#     ])

tt = line.get_table()
line.configure_bend_model(core='adaptive', edge='full')
# line.configure_bend_model(core='bend-kick-bend', edge='full')
# line.configure_bend_model(core='rot-kick-rot', edge='full', num_multipole_kicks=10)

tw = line.twiss(method='4d')

mad.input('twiss, chrom, table=twchr;')
twmad_ch = mad.table.twchr
mad.input('twiss, chrom=false, table=twnochr;')
twmad_nc = mad.table.twnochr

delta_chrom = 1e-4
mad.input(f'''
  ptc_create_universe;
  !ptc_create_layout, time=false, model=2, exact=true, method=6, nst=10;
  ptc_create_layout, time=false, model=1, exact=true, method=6, nst=10;
    select, flag=ptc_twiss, clear;
    select, flag=ptc_twiss, column=name,keyword,s,l,x,px,y,py,beta11,beta22,disp1,k1l;
    ptc_twiss, closed_orbit, icase=56, no=2, deltap=0, table=ptc_twiss,
               summary_table=ptc_twiss_summary, slice_magnets=true;
    ptc_twiss, closed_orbit, icase=56, no=2, deltap={delta_chrom:e}, table=ptc_twiss_pdp,
               summary_table=ptc_twiss_summary_pdp, slice_magnets=true;
    ptc_twiss, closed_orbit, icase=56, no=2, deltap={-delta_chrom:e}, table=ptc_twiss_mdp,
               summary_table=ptc_twiss_summary_mdp, slice_magnets=true;
  ptc_end;
''')

qx_ptc = mad.table.ptc_twiss.mu1[-1]
qy_ptc = mad.table.ptc_twiss.mu2[-1]
dq1_ptc = (mad.table.ptc_twiss_pdp.mu1[-1] - mad.table.ptc_twiss_mdp.mu1[-1]) / (2 * delta_chrom)
dq2_ptc = (mad.table.ptc_twiss_pdp.mu2[-1] - mad.table.ptc_twiss_mdp.mu2[-1]) / (2 * delta_chrom)

ddq1_ptc = (mad.table.ptc_twiss_pdp.mu1[-1] + mad.table.ptc_twiss_mdp.mu1[-1]
            - 2 * mad.table.ptc_twiss.mu1[-1]) / delta_chrom**2
ddq2_ptc = (mad.table.ptc_twiss_pdp.mu2[-1] + mad.table.ptc_twiss_mdp.mu2[-1]
            - 2 * mad.table.ptc_twiss.mu2[-1]) / delta_chrom**2

tptc = mad.table.ptc_twiss
tptc_p = mad.table.ptc_twiss_pdp
tptc_m = mad.table.ptc_twiss_mdp

fp = 1 + delta_chrom
fm = 1 - delta_chrom

# The MAD-X PTC interface rescales the beta functions with (1 + deltap)
# see: https://github.com/MethodicalAcceleratorDesign/MAD-X/blob/eb495b4f926db53f3cd05133638860f910f42fe2/src/madx_ptc_twiss.f90#L1982
# We need to undo that
beta11_p = tptc_p.beta11 / fp
beta11_m = tptc_m.beta11 / fm
beta22_p = tptc_p.beta22 / fp
beta22_m = tptc_m.beta22 / fm
alfa11_p = tptc_p.alfa11
alfa11_m = tptc_m.alfa11
alfa22_p = tptc_p.alfa22
alfa22_m = tptc_m.alfa22

dx_ptc = (tptc_p.x - tptc_m.x) / (2 * delta_chrom)
dy_ptc = (tptc_p.y - tptc_m.y) / (2 * delta_chrom)

betx = 0.5 * (beta11_p + beta11_m)
bety = 0.5 * (beta22_p + beta22_m)
alfx = 0.5 * (alfa11_p + alfa11_m)
alfy = 0.5 * (alfa22_p + alfa22_m)
d_betx = (beta11_p - beta11_m) / (2 * delta_chrom)
d_bety = (beta22_p - beta22_m) / (2 * delta_chrom)
d_alfx = (alfa11_p - alfa11_m) / (2 * delta_chrom)
d_alfy = (alfa22_p - alfa22_m) / (2 * delta_chrom)

bx_ptc = d_betx / betx
by_ptc = d_bety / bety
ax_ptc = d_alfx - d_betx * alfx / betx
ay_ptc = d_alfy - d_bety * alfy / bety
wx_ptc = np.sqrt(ax_ptc**2 + bx_ptc**2)
wy_ptc = np.sqrt(ay_ptc**2 + by_ptc**2)

mad.input('''
  ptc_create_universe;
  ptc_create_layout, time=false, model=1, exact=true, method=6, nst=10;
  ptc_twiss, closed_orbit, icase=56, no=2, deltap=0, table=ptc_twiss,
              summary_table=ptc_twiss_summary, slice_magnets=true;
''')
delta_arr = np.linspace(-0.7e-4, 0.7e-4, 5)
tw_arr_ptc = []
for delta_chrom in delta_arr:
    mad.input(f'''

        select, flag=ptc_twiss, clear;
        select, flag=ptc_twiss, column=name,keyword,s,l,x,px,y,py,beta11,beta22,disp1,k1l;

        ptc_twiss, closed_orbit, icase=56, no=2, deltap={delta_chrom:e}, table=ptc_twiss_pdp,
                summary_table=ptc_twiss_summary_pdp, slice_magnets=true;
    ''')
    tw_arr_ptc.append(mad.table.ptc_twiss_pdp.dframe())

mad.input('''
  ptc_end;
''')

# Same for xsuite
tw_arr_xs = []
for delta_chrom in delta_arr:
    tw_arr_xs.append(line.twiss(method='4d', delta0=delta_chrom))

qx_ptc = mad.table.ptc_twiss.mu1[-1]
qy_ptc = mad.table.ptc_twiss.mu2[-1]

qx_vs_delta_ptc = np.array([tt.mu1[-1] for tt in tw_arr_ptc])
qy_vs_delta_ptc = np.array([tt.mu2[-1] for tt in tw_arr_ptc])

qx_vs_delta_xs = np.array([tt.mux[-1] for tt in tw_arr_xs])
qy_vs_delta_xs = np.array([tt.muy[-1] for tt in tw_arr_xs])

print(f'qx xsuite:          {tw.qx}')
print(f'qx ptc:             {qx_ptc}')
print(f'qx mad (chrom=F):   {twmad_nc.summary.q1}')
print()
print(f'qy xsuite:          {tw.qy}')
print(f'qy ptc:             {qy_ptc}')
print(f'qy mad (chrom=F):   {twmad_nc.summary.q2}')
print()
print(f'dqx xsuite:         {tw.dqx}')
print(f'dqx ptc:            {dq1_ptc}')
print(f'dqx mad (chrom=F):  {twmad_nc.summary.dq1 * seq.beam.beta}')
print()
print(f'dqy xsuite:         {tw.dqy}')
print(f'dqy ptc:            {dq2_ptc}')
print(f'dqy mad (chrom=F):  {twmad_nc.summary.dq2 * seq.beam.beta}')


import matplotlib.pyplot as plt
plt.close('all')
figsize = (6.4*1.3, 4.8*1.3)

fig_abx = plt.figure(101, figsize=figsize)

ax1 = plt.subplot(2,1,1)
plt.plot(tw.s, tw.ax_chrom, label='xsuite')
plt.plot(tptc.s, ax_ptc, '--', label='ptc')
plt.ylabel(r'$A_x$')
plt.legend(loc='best')

ax2 = plt.subplot(2,1,2, sharex=ax1)
plt.plot(tw.s, tw.bx_chrom)
plt.plot(tptc.s, bx_ptc, '--')
plt.ylabel(r'$B_x$')

fig_aby = plt.figure(111, figsize=figsize)
ax3 = plt.subplot(2,1,1, sharex=ax1)
plt.plot(tw.s, tw.ay_chrom)
plt.plot(tptc.s, ay_ptc, '--')
plt.ylabel(r'$A_y$')
plt.xlabel('s [m]')

ax4 = plt.subplot(2,1,2, sharex=ax1)
plt.plot(tw.s, tw.by_chrom)
plt.plot(tptc.s, by_ptc, '--')
plt.ylabel(r'$B_y$')
plt.xlabel('s [m]')

# Same for beta and Wxy
fig_bet = plt.figure(102, figsize=figsize)

ax21 = plt.subplot(2,1,1, sharex=ax1)
plt.plot(tw.s, tw.betx, label='xsuite')
plt.plot(tptc.s, tptc.beta11, '--', label='ptc')
# plt.plot(twmad_ch.s, twmad_ch.betx, 'r--', label='mad')
plt.ylabel(r'$\beta_x$')
plt.legend(loc='best')

ax22 = plt.subplot(2,1,2, sharex=ax1)
plt.plot(tw.s, tw.bety)
plt.plot(tptc.s, tptc.beta22, '--')
# plt.plot(twmad_ch.s, twmad_ch.bety, 'r--')
plt.ylabel(r'$\beta_y$')
plt.xlabel('s [m]')

fig_w = plt.figure(112, figsize=figsize)
ax23 = plt.subplot(2,1,1, sharex=ax1)
plt.plot(tw.s, tw.wx_chrom)
plt.plot(tptc.s, wx_ptc, '--')
# plt.plot(twmad_ch.s, twmad_ch.wx * tw.beta0, 'r--')
plt.ylabel(r'$W_x$')

ax24 = plt.subplot(2,1,2, sharex=ax1)
plt.plot(tw.s, tw.wy_chrom)
plt.plot(tptc.s, wy_ptc, '--')
# plt.plot(twmad_ch.s, twmad_ch.wy * tw.beta0, 'r--')
plt.ylabel(r'$W_y$')
plt.xlabel('s [m]')

# Same for orbit
fig_co = plt.figure(103, figsize=figsize)
ax31 = plt.subplot(2,1,1, sharex=ax1)
plt.plot(tw.s, tw.x * 1e3, label='xsuite')
plt.plot(tptc.s, tptc.x * 1e3, '--', label='ptc')
plt.ylabel('x [mm]')

ax32 = plt.subplot(2,1,2, sharex=ax1)
plt.plot(tw.s, tw.y * 1e3)
plt.plot(tptc.s, tptc.y * 1e3, '--')
plt.ylabel('y [mm]')
plt.xlabel('s [m]')

# Same for dispersion
fig_disp = plt.figure(104, figsize=figsize)
ax41 = plt.subplot(2,1,1, sharex=ax1)
plt.plot(tw.s, tw.dx, label='xsuite')
plt.plot(tptc.s, dx_ptc, '--', label='ptc')
plt.ylabel(r'$D_x$ [m]')

ax42 = plt.subplot(2,1,2, sharex=ax1)
plt.plot(tw.s, tw.dy)
plt.plot(tptc.s, dy_ptc, '--')
plt.ylabel(r'$D_y$ [m]')
plt.xlabel('s [m]')

tt = line.get_table()
tbends = tt.rows[tt.element_type == 'Bend']
tquads = tt.rows[tt.element_type == 'Quadrupole']
tcombined = tt.rows[tt.element_type == 'CombinedFunctionMagnet']
for ax in [ax1, ax2, ax3, ax4, ax21, ax22, ax23, ax24, ax31, ax32, ax41, ax42]:
    for nn in tbends.name:
        ax.axvspan(tbends['s', nn], tbends['s', nn] + line[nn].length, color='b',
                    alpha=0.2, lw=0)
    for nn in tquads.name:
        ax.axvspan(tquads['s', nn], tquads['s', nn] + line[nn].length, color='r',
                    alpha=0.2, lw=0)
    for nn in tcombined.name:
        ax.axvspan(tcombined['s', nn], tcombined['s', nn] + line[nn].length, color='g',
                    alpha=0.2, lw=0)


plt.figure(2, figsize=(6.4*1.4, 4.8*1.4))
ax1 = plt.subplot(211)
plt.plot(delta_arr*1e3, qx_vs_delta_ptc - dq1_ptc * delta_arr - qx_ptc, label='ptc')
plt.plot(delta_arr*1e3, qx_vs_delta_xs - tw.dqx * delta_arr - tw.qx, label='xsuite')
plt.ylabel(r'$\Delta Q_x$')
plt.legend()

ax2 = plt.subplot(212, sharex=ax1)
plt.plot(delta_arr*1e3, qy_vs_delta_ptc - dq2_ptc * delta_arr - qy_ptc)
plt.plot(delta_arr*1e3, qy_vs_delta_xs - tw.dqy * delta_arr - tw.qy)
plt.xlabel(r'$\delta$ [1e-3]')
plt.ylabel(r'$\Delta Q_y$')

plt.suptitle(f"PTC:    Qx = {qx_ptc:.4f}, Qy = {qy_ptc:.4f} Q'x={dq1_ptc:.4f}, Q'y={dq2_ptc:.4f}\n"
             f"Xsuite: Qx = {tw.qx:.4f}, Qy = {tw.qy:.4f} Q'x={tw.dqx:.4f}, Q'y={tw.dqy:.4f}\n"
             "Linear chromaticity removed from the plot")

title = f'{machine.upper()}' + r' - $\beta_0$' + f' = {tw.beta0:.4f}'
for fig in [fig_abx, fig_aby, fig_bet, fig_w, fig_co, fig_disp]:
    fig.suptitle(title)

plt.show()


from cpymad.madx import Madx
import xtrack as xt

mad = Madx()

mad.input('''

    k1=0.2;
    l = 5.;

    radius = 0.3;

    dknr0=0.02;
    dknr1=-0.04;
    dknr2=0.06;
    dkns3=0.2;

    dksr0=0.03;
    dksr1=-0.05;
    dksr2=0.07;
    dksr3=0.3;

    elm: quadrupole, k1:=k1, l=5.;

    seq: sequence, l=50;
    elm1:elm, at=5;
    elm2:elm, at=10;
    elm3:elm, at=20;
    endsequence;

    beam;
    use,sequence=seq;

    select,pattern=elm,flag=error;
    efcomp, radius=radius,
        dknr={dknr0,dknr1,dknr2,dknr3},
        dksr={dksr0,dksr1,dksr2,dksr3}
        ;

''')

line = xt.Line.from_madx_sequence(mad.sequence.seq,
                                  enable_field_errors=True,
                                  allow_thick=True)

# Remember to set num_multipole_kicks







import xtrack as xt

import xtrack._temp.lhc_match as lm

collider = xt.Environment.from_json('hllhc.json')
collider.build_trackers()
collider.vars.load_madx_optics_file(
    "../../test_data/hllhc15_thick/opt_round_150_1500.madx")

for aa in lm.ARC_NAMES:
    print(f"Matching arc {aa}")
    target_mux_b1 = collider.varval[f'mux{aa}b1']
    target_muy_b1 = collider.varval[f'muy{aa}b1']
    target_mux_b2 = collider.varval[f'mux{aa}b2']
    target_muy_b2 = collider.varval[f'muy{aa}b2']

    opt = lm.match_arc_phase_advance(collider=collider, arc_name=aa,
                    target_mux_b1=target_mux_b1, target_muy_b1=target_muy_b1,
                    target_mux_b2=target_mux_b2, target_muy_b2=target_muy_b2)

# Check
for aa in lm.ARC_NAMES:
    collider.varval[f'kqtf.a{aa}b1'] *= 1.1
    collider.varval[f'kqtf.a{aa}b2'] *=  0.9
    collider.varval[f'kqtd.a{aa}b1'] *= 0.15
    collider.varval[f'kqtd.a{aa}b2'] *= 1.15
    collider.varval[f'kqd.a{aa}'] *= 1.03
    collider.varval[f'kqf.a{aa}'] *= 1.02

for aa in lm.ARC_NAMES:
    print(f"Matching arc {aa}")
    target_mux_b1 = collider.varval[f'mux{aa}b1']
    target_muy_b1 = collider.varval[f'muy{aa}b1']
    target_mux_b2 = collider.varval[f'mux{aa}b2']
    target_muy_b2 = collider.varval[f'muy{aa}b2']

    opt = lm.match_arc_phase_advance(collider=collider, arc_name=aa,
                    target_mux_b1=target_mux_b1, target_muy_b1=target_muy_b1,
                    target_mux_b2=target_mux_b2, target_muy_b2=target_muy_b2)


import xtrack as xt
from cpymad.madx import Madx

# Load a very simple sequence from MAD-X
mad = Madx()
mad.input("""
    seq: sequence, l=4;
    b1: sbend, at=0.5, angle=0.2, l=1;
    b2: sbend, at=2.5, angle=0.3, l=1;
    endsequence;

    beam;
    use,sequence=seq;
""")

line = xt.Line.from_madx_sequence(mad.sequence.seq)
line.build_tracker()

print('The line as imported from MAD-X:')
print(line.get_table())

# Shift and tilt selected elements
line['b1'].shift_x = -0.1
line['b1'].rot_s_rad = 0.8
line['b2'].shift_s= 0.2
line['b2'].rot_s_rad = -0.8

tt = line.get_table(attr=True)
tt.cols['s', 'element_type', 'isthick', 'shift_x', 'shift_y', 'shift_s', 'rot_s_rad']
# returns:
#
# name       s element_type isthick shift_x shift_y shift_s rot_s_rad
# seq$start  0 Marker         False       0       0       0         0
# b1         0 Bend            True    -0.1       0       0       0.8
# drift_0    1 Drift           True       0       0       0         0
# b2         2 Bend            True       0       0     0.2      -0.8
# drift_1    3 Drift           True       0       0       0         0
# seq$end    4 Marker         False       0       0       0         0
# _end_point 4                False       0       0       0         0

# Slice the line
slicing_strategies = [
    xt.Strategy(slicing=None),  # Default catch-all
    xt.Strategy(slicing=xt.Teapot(2), element_type=xt.Bend),
]
line.slice_thick_elements(slicing_strategies)
line.build_tracker()

# Inspect
tt = line.get_table(attr=True)
tt.cols['s', 'element_type', 'isthick', 'parent_name',
        'shift_x', 'shift_y', 'shift_s', 'rot_s_rad']
# returns:
#
# Table: 23 rows, 9 cols
# name                 s element_type       isthick parent_name shift_x shift_y shift_s rot_s_rad
# seq$start            0 Marker               False        None       0       0       0         0
# b1_entry             0 Marker               False        None       0       0       0         0
# b1..entry_map        0 ThinSliceBendEntry   False          b1    -0.1       0       0       0.8
# drift_b1..0          0 DriftSliceBend        True          b1       0       0       0         0
# b1..0         0.166667 ThinSliceBend        False          b1    -0.1       0       0       0.8
# drift_b1..1   0.166667 DriftSliceBend        True          b1       0       0       0         0
# b1..1         0.833333 ThinSliceBend        False          b1    -0.1       0       0       0.8
# drift_b1..2   0.833333 DriftSliceBend        True          b1       0       0       0         0
# b1..exit_map         1 ThinSliceBendExit    False          b1    -0.1       0       0       0.8
# b1_exit              1 Marker               False        None       0       0       0         0
# drift_0              1 Drift                 True        None       0       0       0         0
# b2_entry             2 Marker               False        None       0       0       0         0
# b2..entry_map        2 ThinSliceBendEntry   False          b2       0       0     0.2      -0.8
# drift_b2..0          2 DriftSliceBend        True          b2       0       0       0        -0
# b2..0          2.16667 ThinSliceBend        False          b2       0       0     0.2      -0.8
# drift_b2..1    2.16667 DriftSliceBend        True          b2       0       0       0        -0
# b2..1          2.83333 ThinSliceBend        False          b2       0       0     0.2      -0.8
# drift_b2..2    2.83333 DriftSliceBend        True          b2       0       0       0        -0
# b2..exit_map         3 ThinSliceBendExit    False          b2       0       0     0.2      -0.8
# b2_exit              3 Marker               False        None       0       0       0         0
# drift_1              3 Drift                 True        None       0       0       0         0
# seq$end              4 Marker               False        None       0       0       0         0
# _end_point           4                      False        None       0       0       0         0


# Update misalignment for one element. We act on the parent and the effect is
# propagated to the slices.
line['b2'].rot_s_rad = 0.3
line['b2'].shift_x = 2e-3

# Inspect
tt = line.get_table(attr=True)
tt.cols['s', 'element_type', 'isthick', 'parent_name', 'shift_x', 'shift_y', 'rot_s_rad']
# returns:
#
# Table: 23 rows, 8 cols
# name                 s element_type       isthick parent_name shift_x shift_y rot_s_rad
# seq$start            0 Marker               False        None       0       0         0
# b1_entry             0 Marker               False        None       0       0         0
# b1..entry_map        0 ThinSliceBendEntry   False          b1    -0.1       0       0.8
# drift_b1..0          0 DriftSliceBend        True          b1       0       0         0
# b1..0         0.166667 ThinSliceBend        False          b1    -0.1       0       0.8
# drift_b1..1   0.166667 DriftSliceBend        True          b1       0       0         0
# b1..1         0.833333 ThinSliceBend        False          b1    -0.1       0       0.8
# drift_b1..2   0.833333 DriftSliceBend        True          b1       0       0         0
# b1..exit_map         1 ThinSliceBendExit    False          b1    -0.1       0       0.8
# b1_exit              1 Marker               False        None       0       0         0
# drift_0              1 Drift                 True        None       0       0         0
# b2_entry             2 Marker               False        None       0       0         0
# b2..entry_map        2 ThinSliceBendEntry   False          b2   0.002       0       0.3
# drift_b2..0          2 DriftSliceBend        True          b2       0       0         0
# b2..0          2.16667 ThinSliceBend        False          b2   0.002       0       0.3
# drift_b2..1    2.16667 DriftSliceBend        True          b2       0       0         0
# b2..1          2.83333 ThinSliceBend        False          b2   0.002       0       0.3
# drift_b2..2    2.83333 DriftSliceBend        True          b2       0       0         0
# b2..exit_map         3 ThinSliceBendExit    False          b2   0.002       0       0.3
# b2_exit              3 Marker               False        None       0       0         0
# drift_1              3 Drift                 True        None       0       0         0
# seq$end              4 Marker               False        None       0       0         0
# _end_point           4                      False        None       0       0         0


# copyright ############################### #
# This file is part of the Xtrack Package.  #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

# MADX Reference:
# https://github.com/MethodicalAcceleratorDesign/MAD-X/blob/2dcd046b1f6ca2b44ef67c8d572ff74370deee25/src/survey.f90


import numpy as np

from xdeps import Table
import xtrack as xt

# Required functions
# ==================================================
def get_w_from_angles(theta, phi, psi, reverse_xs=False):
    """W matrix, see MAD-X manual"""
    costhe = np.cos(theta)
    cosphi = np.cos(phi)
    cospsi = np.cos(psi)
    sinthe = np.sin(theta)
    sinphi = np.sin(phi)
    sinpsi = np.sin(psi)
    w = np.zeros([3, 3])
    w[0, 0] = +costhe * cospsi - sinthe * sinphi * sinpsi
    w[0, 1] = -costhe * sinpsi - sinthe * sinphi * cospsi
    w[0, 2] = sinthe * cosphi
    w[1, 0] = cosphi * sinpsi
    w[1, 1] = cosphi * cospsi
    w[1, 2] = sinphi
    w[2, 0] = -sinthe * cospsi - costhe * sinphi * sinpsi
    w[2, 1] = +sinthe * sinpsi - costhe * sinphi * cospsi
    w[2, 2] = costhe * cosphi

    if reverse_xs:
        w[:, 0] *= -1
        w[:, 2] *= -1

    return w


def get_angles_from_w(w, reverse_xs=False):
    """Inverse function of get_w_from_angles()"""
    # w[0, 2]/w[2, 2] = (sinthe * cosphi)/(costhe * cosphi)
    # w[1, 0]/w[1, 1] = (cosphi * sinpsi)/(cosphi * cospsi)
    # w[1, 2]/w[1, 1] = (sinphi)/(cosphi * cospsi)

    if reverse_xs:
        w = w.copy()
        w[:, 0] *= -1
        w[:, 2] *= -1

    theta = np.arctan2(w[0, 2], w[2, 2])
    psi = np.arctan2(w[1, 0], w[1, 1])
    phi = np.arctan2(w[1, 2], w[1, 1] / np.cos(psi))

    # TODO: arctan2 returns angle between [-pi,pi]. Hence theta ends up not at 2pi after a full survey
    return theta, phi, psi


def advance_bend(v, w, R, S):
    """Advancing through bending element, see MAD-X manual:
    v2 = w1*R + v1  | w2 = w1*S"""
    return np.dot(w, R) + v, np.dot(w, S)


def advance_drift(v, w, R):
    """Advancing through drift element, see MAD-X manual:
    v2 = w1*R + v1  | w2 = w1*S -> S is unity"""
    return np.dot(w, R) + v, w


def advance_element(v, w, length=0, angle=0, tilt=0):
    """Computing the advance element-by-element. See MAD-X manual for generation of R and S"""
    if angle == 0:
        R = np.array([0, 0, length])
        return advance_drift(v, w, R)
    elif tilt == 0:
        # Relevant sine/cosine
        ca = np.cos(angle)
        sa = np.sin(angle)
        # ------
        rho = length / angle
        R = np.array([rho * (ca - 1), 0, rho * sa])
        S = np.array([[ca, 0, -sa], [0, 1, 0], [sa, 0, ca]])
        return advance_bend(v, w, R, S)

    else:
        # Relevant sine/cosine
        ca = np.cos(angle)
        sa = np.sin(angle)
        ct = np.cos(tilt)
        st = np.sin(tilt)
        # ------
        rho = length / angle
        R = np.array([rho * (ca - 1), 0, rho * sa])
        S = np.array([[ca, 0, -sa], [0, 1, 0], [sa, 0, ca]])

        # Orthogonal rotation matrix for tilt
        T = np.array([[ct, -st, 0], [st, ct, 0], [0, 0, 1]])
        Tinv = np.array([[ct, st, 0], [-st, ct, 0], [0, 0, 1]])

        return advance_bend(v, w, np.dot(T, R), np.dot(T, np.dot(S, Tinv)))


class SurveyTable(Table):

    _error_on_row_not_found = True

    def reverse(self, X0=None, Y0=None, Z0=None, theta0=None,
                phi0=None, psi0=None, element0=None):

        if element0 is None:
            element0 = len(self.name) - self.element0 - 1

        if (X0 is not None or Y0 is not None or Z0 is not None
                or theta0 is not None or phi0 is not None or psi0 is not None):
            assert (X0 is not None and Y0 is not None and Z0 is not None
                and theta0 is not None and phi0 is not None and psi0 is not None
                    ), (
            "X0, Y0, Z0, theta0, phi0, psi0 must be all None or all not None")

        if X0 is None:
            X0 = self.X[self.element0]
            Y0 = self.Y[self.element0]
            Z0 = self.Z[self.element0]
            theta0 = self.theta[self.element0]
            phi0 = self.phi[self.element0]
            psi0 = self.psi[self.element0]

        # We cut away the last marker (added by survey) and reverse the order
        out_drift_length = list(self.drift_length[:-1][::-1])
        out_angle = list(-self.angle[:-1][::-1])
        out_tilt = list(-self.tilt[:-1][::-1])
        out_name = list(self.name[:-1][::-1])
        out_element_type = list(self.element_type[:-1][::-1])
        out_isthick = list(self.isthick[:-1][::-1])
        out_length = list(self.length[:-1][::-1])

        if type(element0) is str:
            element0 = out_name.index(element0)

        X, Y, Z, theta, phi, psi = compute_survey(
                                        X0, Y0, Z0, theta0, phi0, psi0,
                                        out_drift_length, out_angle, out_tilt,
                                        element0=element0)

        # Initializing dictionary
        out_columns = {}
        out_columns["X"] = np.array(X)
        out_columns["Y"] = np.array(Y)
        out_columns["Z"] = np.array(Z)
        out_columns["theta"] = np.unwrap(theta)
        out_columns["phi"] = np.unwrap(phi)
        out_columns["psi"] = np.unwrap(psi)

        out_columns["name"] = np.array(list(out_name) + ["_end_point"])
        out_columns["s"] = self.s[-1] - self.s[::-1]

        out_columns['drift_length'] = np.array(out_drift_length + [0.])
        out_columns['angle'] = np.array(out_angle + [0.])
        out_columns['tilt'] = np.array(out_tilt + [0.])
        out_columns['element_type'] = np.array(out_element_type + [""])
        out_columns['isthick'] = np.array(out_isthick + [False])
        out_columns['length'] = np.array(out_length + [0.])

        out_scalars = {}
        out_scalars["element0"] = element0

        out = SurveyTable(data=(out_columns | out_scalars),
                          col_names=out_columns.keys())

        return out

    def plot(self, element_width=None, legend=True, **kwargs):
        # Shallow copy of self
        out_sv_table = SurveyTable.__new__(SurveyTable)
        out_sv_table.__dict__.update(self.__dict__)
        out_sv_table._data = self._data.copy()

        # Removing the count for repeated elements
        out_sv_table.name = np.array([nn.split('::')[0] for nn in out_sv_table.name])

        if element_width is None:
            x_range = max(self.X) - min(self.X)
            y_range = max(self.Y) - min(self.Y)
            z_range = max(self.Z) - min(self.Z)
            element_width = max([x_range, y_range, z_range]) * 0.03
        import xplt
        xplt.FloorPlot(out_sv_table, element_width=element_width, **kwargs)
        if legend:
            import matplotlib.pyplot as plt
            plt.legend()


# ==================================================
# Main function
# ==================================================
def survey_from_line(line, X0=0, Y0=0, Z0=0, theta0=0, phi0=0, psi0=0,
                        element0=0,
                        values_at_element_exit=False,
                        reverse=True):
    """Execute SURVEY command. Based on MADX equivalent.
    Attributes, must be given in this order in the dictionary:
    X0        (float)    Initial X position in meters.
    Y0        (float)    Initial Y position in meters.
    Z0        (float)    Initial Z position in meters.
    theta0    (float)    Initial azimuthal angle in radians.
    phi0      (float)    Initial elevation angle in radians.
    psi0      (float)    Initial roll angle in radians."""

    if reverse:
        raise ValueError('`survey(..., reverse=True)` not supported anymore. '
                         'Use `survey(...).reverse()` instead.')

    assert not values_at_element_exit, "Not implemented yet"

    # Extract angle and tilt from elements
    tt = line.get_table(attr = True)
    angle = tt.angle_rad
    tilt = tt.rot_s_rad
    drift_length = tt.length.copy()
    drift_length[~tt.isthick] = 0

    if type(element0) == str:
        element0 = line.element_names.index(element0)

    X, Y, Z, theta, phi, psi = compute_survey(
        X0, Y0, Z0, theta0, phi0, psi0, drift_length[:-1], angle[:-1], tilt[:-1],
        element0=element0)

    # Initializing dictionary
    out_columns = {}
    out_scalars = {}
    out_columns["X"] = np.array(X)
    out_columns["Y"] = np.array(Y)
    out_columns["Z"] = np.array(Z)
    out_columns["theta"] = np.unwrap(theta)
    out_columns["phi"] = np.unwrap(phi)
    out_columns["psi"] = np.unwrap(psi)

    out_columns["name"] = tt.name
    out_columns["s"] = tt.s
    out_columns["length"] = tt.length
    out_columns["isthick"] = tt.isthick
    out_columns['drift_length'] = drift_length
    out_columns['angle'] = angle
    out_columns['tilt'] = tilt
    out_columns['element_type'] = tt.element_type

    out_scalars['element0'] = element0

    out = SurveyTable(data={**out_columns, **out_scalars},  # this is a merge
                      col_names=out_columns.keys())
    out._data['line'] = line

    return out


def compute_survey(X0, Y0, Z0, theta0, phi0, psi0, drift_length, angle, tilt,
                   element0=0, reverse_xs=False):

    if element0 != 0:
        assert not(reverse_xs), "Not implemented yet"
        drift_forward = drift_length[element0:]
        angle_forward = angle[element0:]
        tilt_forward = tilt[element0:]
        (X_forward, Y_forward, Z_forward, theta_forward, phi_forward,
            psi_forward) = compute_survey(X0, Y0, Z0, theta0, phi0, psi0,
                                    drift_forward, angle_forward, tilt_forward)

        drift_backward = drift_length[:element0][::-1]
        angle_backward = -np.array(angle[:element0][::-1])
        tilt_backward = -np.array(tilt[:element0][::-1])
        (X_backward, Y_backward, Z_backward, theta_backward, phi_backward,
            psi_backward) = compute_survey(X0, Y0, Z0, theta0, phi0, psi0,
                                    drift_backward, angle_backward, tilt_backward,
                                    reverse_xs=True)

        X = np.array(X_backward[::-1][:-1] + X_forward)
        Y = np.array(Y_backward[::-1][:-1] + Y_forward)
        Z = np.array(Z_backward[::-1][:-1] + Z_forward)
        theta = np.array(theta_backward[::-1][:-1] + theta_forward)
        phi = np.array(phi_backward[::-1][:-1] + phi_forward)
        psi = np.array(psi_backward[::-1][:-1]+ psi_forward)
        return X, Y, Z, theta, phi, psi

    X = []
    Y = []
    Z = []
    theta = []
    phi = []
    psi = []
    v = np.array([X0, Y0, Z0])
    w = get_w_from_angles(theta=theta0, phi=phi0, psi=psi0,
                          reverse_xs=reverse_xs)
    # Advancing element by element
    for ll, aa, tt in zip(drift_length, angle, tilt):

        th, ph, ps = get_angles_from_w(w, reverse_xs=reverse_xs)

        X.append(v[0])
        Y.append(v[1])
        Z.append(v[2])
        theta.append(th)
        phi.append(ph)
        psi.append(ps)

        # Advancing
        v, w = advance_element(v, w, length=ll, angle=aa, tilt=tt)

    # Last marker
    th, ph, ps = get_angles_from_w(w, reverse_xs=reverse_xs)
    X.append(v[0])
    Y.append(v[1])
    Z.append(v[2])
    theta.append(th)
    phi.append(ph)
    psi.append(ps)

    # Returns as SurveyTable object
    return X, Y, Z, theta, phi, psi



import xtrack as xt
from cpymad.madx import Madx

import numpy as np

mad = Madx()

tilt_deg = 12
shift_x = -2e-3
shift_y = 3e-3
k1 = 0.2
tilt_rad = np.deg2rad(tilt_deg)

x_test = 1e-3
px_test = 2e-3
y_test = 3e-3
py_test = 4e-3

mad.input(f"""
k1={k1};

elm: quadrupole,
    k1:=k1,
    l=1,
    tilt={tilt_rad};

seq: sequence, l=1;
elm: elm, at=0.5;
endsequence;

beam, particle=proton, gamma=100;
use, sequence=seq;

select,flag=error,clear;
select,flag=error,pattern=elm;
ealign, dx={shift_x}, dy={shift_y};

twiss, betx=1, bety=1, x={x_test}, px={px_test}, y={y_test}, py={py_test};

""")

elm = xt.Quadrupole(k1=k1, length=1)

elm_tilted = xt.Quadrupole(k1=k1, length=1, rot_s_rad=tilt_rad,
                           shift_x=shift_x, shift_y=shift_y)

lsandwitch = xt.Line(elements=[
    xt.XYShift(dx=shift_x, dy=shift_y),
    xt.SRotation(angle=tilt_deg),
    elm,
    xt.SRotation(angle=-tilt_deg),
    xt.XYShift(dx=-shift_x, dy=-shift_y)
])
lsandwitch.build_tracker()

l_tilted = xt.Line(elements=[elm_tilted])
l_tilted.build_tracker()

lmad = xt.Line.from_madx_sequence(mad.sequence.seq, enable_align_errors=True)
lmad.build_tracker()

p0 = xt.Particles(x=x_test, px=px_test, y=y_test, py=py_test, gamma0=100)

pmad = p0.copy()
lmad.track(pmad)

psandwitch = p0.copy()
lsandwitch.track(psandwitch)

plinetilted = p0.copy()
l_tilted.track(plinetilted)

peletitled = p0.copy()
elm_tilted.track(peletitled)

assert elm.rot_s_rad == 0
elm.rot_s_rad = tilt_rad
elm.shift_x = shift_x
elm.shift_y = shift_y
pprop = p0.copy()
elm.track(pprop)

assert_allclose = np.testing.assert_allclose
for pp in [psandwitch, plinetilted, pmad, peletitled, pprop]:
    assert_allclose(pp.x, mad.table.twiss.x[-1], rtol=0, atol=1e-12)
    assert_allclose(pp.px, mad.table.twiss.px[-1], rtol=0, atol=1e-12)
    assert_allclose(pp.y, mad.table.twiss.y[-1], rtol=0, atol=1e-12)
    assert_allclose(pp.py, mad.table.twiss.py[-1], rtol=0, atol=1e-12)
    assert_allclose(pp.zeta, pp.beta0[0]*mad.table.twiss.t[-1], rtol=0, atol=1e-12)



import numpy as np
import pytest
from scipy.constants import c as clight
from scipy.constants import e as qe

import xobjects as xo
import xtrack as xt


def test_multispecies_multipole():

    p_ref1 = xt.Particles(
        mass0=xt.PROTON_MASS_EV,
        q0=2,
        gamma0=1.2)

    p_ref2 = xt.Particles(
        mass0=2*xt.PROTON_MASS_EV,
        q0=3,
        gamma0=1.5)

    # Build a particle referred to reference 1
    p1 = p_ref1.copy()
    p1.x = 1e-2
    p1.y = 2e-2
    p1.delta = 0.5
    P_p1 = (1 + p1.delta) * p1.p0c

    # Build the same particle referred to reference 2
    p1_ref2 = p_ref2.copy()
    p1_ref2_mass_ratio = p1.mass0 / p1_ref2.mass0
    p1_ref2_charge_ratio = p1.q0 / p1_ref2.q0

    p1_ref2.x = p1.x
    p1_ref2.y = p1.y
    p1_ref2.zeta = p1.zeta / p1.beta0 * p1_ref2.beta0
    p1_ref2.charge_ratio = p1_ref2_charge_ratio
    p1_ref2.chi = p1_ref2_charge_ratio / p1_ref2_mass_ratio
    p1_ref2.delta = P_p1 / p1_ref2_mass_ratio / p1_ref2.p0c - 1

    xo.assert_allclose(p1_ref2.mass, p1.mass, atol=0, rtol=1e-14)
    xo.assert_allclose(p1_ref2.charge, p1.charge, atol=0, rtol=1e-14)
    xo.assert_allclose(p1_ref2.energy, p1.energy, atol=0, rtol=1e-14)
    xo.assert_allclose(p1_ref2.rvv * p1_ref2.beta0, p1.rvv * p1.beta0, atol=0, rtol=1e-14)

    xo.assert_allclose(p1.rpp, 1 / (1 + p1.delta), atol=0, rtol=1e-14)
    xo.assert_allclose(p1_ref2.rpp, 1 / (1 + p1_ref2.delta), atol=0, rtol=1e-14)

    p1c = p1.p0c / p1.rpp * p1.mass_ratio
    p1c_ref2 = p1_ref2.p0c / p1_ref2.rpp * p1_ref2.mass_ratio
    xo.assert_allclose(p1c, p1c_ref2, atol=0, rtol=1e-14)

    L_bend = 1.

    B_T = 0.5
    BsT = 0.1
    G_Tm = 0.1
    Gs_Tm = -0.05

    P0_J_ref1 = p_ref1.p0c[0] / clight * qe
    h_bend_ref1 = B_T * qe * p_ref1.charge[0] / P0_J_ref1 # This is brho
    theta_bend_ref1 = h_bend_ref1 * L_bend
    theta_skew_ref1 = BsT * qe * p_ref1.charge[0] / P0_J_ref1
    k1l_ref1 = G_Tm * qe * p_ref1.charge[0] / P0_J_ref1
    k1sl_ref1 = Gs_Tm * qe * p_ref1.charge[0] / P0_J_ref1


    P0_J_ref2 = p_ref2.p0c[0] / clight * qe
    h_bend_ref2 = B_T * qe * p_ref2.charge[0] / P0_J_ref2
    theta_bend_ref2 = h_bend_ref2 * L_bend
    theta_skew_ref2 = BsT * qe * p_ref2.charge[0] / P0_J_ref2
    k1l_ref2 = G_Tm * qe * p_ref2.charge[0] / P0_J_ref2
    k1sl_ref2 = Gs_Tm * qe * p_ref2.charge[0] / P0_J_ref2

    n_slices = 100

    dipole_ref1 = xt.Multipole(knl=[theta_bend_ref1/n_slices, k1l_ref1/n_slices],
                            ksl=[theta_skew_ref1/n_slices, k1sl_ref1/n_slices],
                            length=L_bend/n_slices, hxl=0.2/n_slices)
    dipole_ref2 = xt.Multipole(knl=[theta_bend_ref2/n_slices, k1l_ref2/n_slices],
                            ksl=[theta_skew_ref2/n_slices, k1sl_ref2/n_slices],
                            length=L_bend/n_slices, hxl=0.2/n_slices)

    ele_ref1 = []
    for ii in range(n_slices):
        ele_ref1.append(xt.Drift(length=L_bend/n_slices/2))
        ele_ref1.append(dipole_ref1)
        ele_ref1.append(xt.Drift(length=L_bend/n_slices/2))

    ele_ref2 = []
    for ii in range(n_slices):
        ele_ref2.append(xt.Drift(length=L_bend/n_slices/2))
        ele_ref2.append(dipole_ref2)
        ele_ref2.append(xt.Drift(length=L_bend/n_slices/2))

    line_ref1 = xt.Line(elements=ele_ref1)
    line_ref2 = xt.Line(elements=ele_ref2)

    line_ref1.append_element(element=xt.Marker(), name='endmarker')
    line_ref2.append_element(element=xt.Marker(), name='endmarker')

    line_ref1.build_tracker()
    line_ref2.build_tracker()

    line_ref1.track(p1, ele_start=0, ele_stop='endmarker')
    line_ref2.track(p1_ref2, ele_start=0, ele_stop='endmarker')

    xo.assert_allclose(p1.x, p1_ref2.x, atol=0, rtol=1e-14)
    xo.assert_allclose(p1.y, p1_ref2.y, atol=0, rtol=1e-14)
    xo.assert_allclose(p1_ref2.mass, p1.mass, atol=0, rtol=1e-14)
    xo.assert_allclose(p1_ref2.charge, p1.charge, atol=0, rtol=1e-14)
    xo.assert_allclose(p1_ref2.energy, p1.energy, atol=0, rtol=1e-14)
    xo.assert_allclose(p1_ref2.rvv * p1_ref2.beta0, p1.rvv * p1.beta0, atol=0, rtol=1e-14)

    # Check absolute time of arrival
    t0_ref1 = p1.s / (p1.beta0 * clight)           # Absolute reference time of arrival
    t0_ref2 = p1_ref2.s / (p1_ref2.beta0 * clight) # Absolute reference time of arrival
    dt_ref1 = -p1.zeta / (p1.beta0 * clight)           # Arrival time relative to reference
    dt_ref2 = -p1_ref2.zeta / (p1_ref2.beta0 * clight) # Arrival time relative to reference

    xo.assert_allclose(t0_ref1 + dt_ref1, t0_ref2 + dt_ref2, atol=1e-11, rtol=0)

@pytest.mark.parametrize('model', ['expanded', 'bend-kick-bend', 'rot-kick-rot'])
@pytest.mark.parametrize('B_T', [0.4, 0])
@pytest.mark.parametrize('hxl', [0.2, 0])
@pytest.mark.parametrize('G_Tm', [0.1, 0])
@pytest.mark.parametrize('S_Tm2', [0.05, 0])
def test_multispecies_bend(model, B_T, hxl, G_Tm, S_Tm2):
    p_ref1 = xt.Particles(
        mass0=xt.PROTON_MASS_EV,
        q0=2,
        gamma0=1.2)

    p_ref2 = xt.Particles(
        mass0=2*xt.PROTON_MASS_EV,
        q0=3,
        gamma0=1.5)

    # Build a particle referred to reference 1
    p1 = p_ref1.copy()
    p1.x = 1e-3
    p1.y = 2e-3
    p1.delta = 0.5
    P_p1 = (1 + p1.delta) * p1.p0c

    # Build the same particle referred to reference 2
    p1_ref2 = p_ref2.copy()
    p1_ref2_mass_ratio = p1.mass0 / p1_ref2.mass0
    p1_ref2_charge_ratio = p1.q0 / p1_ref2.q0

    p1_ref2.x = p1.x
    p1_ref2.y = p1.y
    p1_ref2.zeta = p1.zeta / p1.beta0 * p1_ref2.beta0
    p1_ref2.charge_ratio = p1_ref2_charge_ratio
    p1_ref2.chi = p1_ref2_charge_ratio / p1_ref2_mass_ratio
    p1_ref2.delta = P_p1 / p1_ref2_mass_ratio / p1_ref2.p0c - 1

    xo.assert_allclose(p1_ref2.mass, p1.mass, atol=0, rtol=1e-14)
    xo.assert_allclose(p1_ref2.charge, p1.charge, atol=0, rtol=1e-14)
    xo.assert_allclose(p1_ref2.energy, p1.energy, atol=0, rtol=1e-14)
    xo.assert_allclose(p1_ref2.rvv * p1_ref2.beta0, p1.rvv * p1.beta0, atol=0, rtol=1e-14)

    xo.assert_allclose(p1.rpp, 1 / (1 + p1.delta), atol=0, rtol=1e-14)
    xo.assert_allclose(p1_ref2.rpp, 1 / (1 + p1_ref2.delta), atol=0, rtol=1e-14)

    p1c = p1.p0c / p1.rpp * p1.mass_ratio
    p1c_ref2 = p1_ref2.p0c / p1_ref2.rpp * p1_ref2.mass_ratio
    xo.assert_allclose(p1c, p1c_ref2, atol=0, rtol=1e-14)

    L_bend = 1.

    P0_J_ref1 = p_ref1.p0c[0] / clight * qe
    h_bend_ref1 = B_T * qe * p_ref1.charge[0] / P0_J_ref1 * L_bend # This is brho
    theta_bend_ref1 = h_bend_ref1 * L_bend
    k1l_ref1 = G_Tm * qe * p_ref1.charge[0] / P0_J_ref1 * L_bend
    k2l_ref1 = S_Tm2 * qe * p_ref1.charge[0] / P0_J_ref1 * L_bend


    P0_J_ref2 = p_ref2.p0c[0] / clight * qe
    h_bend_ref2 = B_T * qe * p_ref2.charge[0] / P0_J_ref2 * L_bend
    theta_bend_ref2 = h_bend_ref2 * L_bend
    k1l_ref2 = G_Tm * qe * p_ref2.charge[0] / P0_J_ref2 * L_bend
    k2l_ref2 = S_Tm2 * qe * p_ref2.charge[0] / P0_J_ref2 * L_bend

    n_slices = 10

    dipole_ref1 = xt.Bend(k0=theta_bend_ref1/L_bend, length=L_bend / n_slices,
                        h=hxl/L_bend, k1=k1l_ref1/n_slices/L_bend,
                        knl=[0, 0, k2l_ref1/n_slices])
    dipole_ref2 = xt.Bend(k0=theta_bend_ref2/L_bend, length=L_bend / n_slices,
                        h=hxl/L_bend, k1=k1l_ref2/n_slices/L_bend,
                        knl=[0, 0, k2l_ref2/n_slices])

    if model == 'expanded':
        dipole_ref1.num_multipole_kicks = 5
        dipole_ref2.num_multipole_kicks = 5

    ele_ref1 = []
    for ii in range(n_slices):
        ele_ref1.append(xt.Drift(length=L_bend/n_slices/2))
        ele_ref1.append(dipole_ref1)
        ele_ref1.append(xt.Drift(length=L_bend/n_slices/2))

    ele_ref2 = []
    for ii in range(n_slices):
        ele_ref2.append(xt.Drift(length=L_bend/n_slices/2))
        ele_ref2.append(dipole_ref2)
        ele_ref2.append(xt.Drift(length=L_bend/n_slices/2))

    line_ref1 = xt.Line(elements=ele_ref1)
    line_ref2 = xt.Line(elements=ele_ref2)

    line_ref1.append_element(element=xt.Marker(), name='endmarker')
    line_ref2.append_element(element=xt.Marker(), name='endmarker')

    line_ref1.configure_bend_model(core=model)
    line_ref2.configure_bend_model(core=model)

    line_ref1.build_tracker()
    line_ref2.build_tracker()

    line_ref1.track(p1, ele_start=0, ele_stop='endmarker')
    line_ref2.track(p1_ref2, ele_start=0, ele_stop='endmarker')

    xo.assert_allclose(p1.x, p1_ref2.x, atol=0, rtol=1e-10)
    xo.assert_allclose(p1.y, p1_ref2.y, atol=0, rtol=1e-10)
    xo.assert_allclose(p1_ref2.mass, p1.mass, atol=0, rtol=1e-14)
    xo.assert_allclose(p1_ref2.charge, p1.charge, atol=0, rtol=1e-14)
    xo.assert_allclose(p1_ref2.energy, p1.energy, atol=0, rtol=1e-14)
    xo.assert_allclose(p1_ref2.rvv * p1_ref2.beta0, p1.rvv * p1.beta0, atol=0, rtol=1e-14)

    # Check absolute time of arrival
    t0_ref1 = p1.s / (p1.beta0 * clight)           # Absolute reference time of arrival
    t0_ref2 = p1_ref2.s / (p1_ref2.beta0 * clight) # Absolute reference time of arrival
    dt_ref1 = -p1.zeta / (p1.beta0 * clight)           # Arrival time relative to reference
    dt_ref2 = -p1_ref2.zeta / (p1_ref2.beta0 * clight) # Arrival time relative to reference

    xo.assert_allclose(t0_ref1 + dt_ref1, t0_ref2 + dt_ref2, atol=1e-11, rtol=0)

@pytest.mark.parametrize('model', ['full', 'linear'])
@pytest.mark.parametrize('side', ['entry', 'exit'])
def test_multispecies_dipole_edge(model, side):

    p_ref1 = xt.Particles(
        mass0=xt.PROTON_MASS_EV,
        q0=2,
        gamma0=1.2)

    p_ref2 = xt.Particles(
        mass0=2*xt.PROTON_MASS_EV,
        q0=3,
        gamma0=1.5)

    # Build a particle referred to reference 1
    p1 = p_ref1.copy()
    p1.x = 1e-3
    p1.y = 2e-3
    p1.delta = 0.5
    P_p1 = (1 + p1.delta) * p1.p0c

    # Build the same particle referred to reference 2
    p1_ref2 = p_ref2.copy()
    p1_ref2_mass_ratio = p1.mass0 / p1_ref2.mass0
    p1_ref2_charge_ratio = p1.q0 / p1_ref2.q0

    p1_ref2.x = p1.x
    p1_ref2.y = p1.y
    p1_ref2.zeta = p1.zeta / p1.beta0 * p1_ref2.beta0
    p1_ref2.charge_ratio = p1_ref2_charge_ratio
    p1_ref2.chi = p1_ref2_charge_ratio / p1_ref2_mass_ratio
    p1_ref2.delta = P_p1 / p1_ref2_mass_ratio / p1_ref2.p0c - 1

    xo.assert_allclose(p1_ref2.mass, p1.mass, atol=0, rtol=1e-14)
    xo.assert_allclose(p1_ref2.charge, p1.charge, atol=0, rtol=1e-14)
    xo.assert_allclose(p1_ref2.energy, p1.energy, atol=0, rtol=1e-14)
    xo.assert_allclose(p1_ref2.rvv * p1_ref2.beta0, p1.rvv * p1.beta0, atol=0, rtol=1e-14)

    xo.assert_allclose(p1.rpp, 1 / (1 + p1.delta), atol=0, rtol=1e-14)
    xo.assert_allclose(p1_ref2.rpp, 1 / (1 + p1_ref2.delta), atol=0, rtol=1e-14)

    p1c = p1.p0c / p1.rpp * p1.mass_ratio
    p1c_ref2 = p1_ref2.p0c / p1_ref2.rpp * p1_ref2.mass_ratio
    xo.assert_allclose(p1c, p1c_ref2, atol=0, rtol=1e-14)

    B_T = 0.4
    e1=0.1

    P0_J_ref1 = p_ref1.p0c[0] / clight * qe
    k_bend_ref1 = B_T * qe * p_ref1.charge[0] / P0_J_ref1 # This is brho

    P0_J_ref2 = p_ref2.p0c[0] / clight * qe
    k_bend_ref2 = B_T * qe * p_ref2.charge[0] / P0_J_ref2

    edge_ref1 = xt.DipoleEdge(k=k_bend_ref1, e1=e1, hgap=0.05, fint=0.5, side=side)
    edge_ref2 = xt.DipoleEdge(k=k_bend_ref2, e1=e1, hgap=0.05, fint=0.5, side=side)

    ele_ref1 = [xt.Drift(length=1), edge_ref1, xt.Drift(length=1)]
    ele_ref2 = [xt.Drift(length=1), edge_ref2, xt.Drift(length=1)]

    line_ref1 = xt.Line(elements=ele_ref1)
    line_ref2 = xt.Line(elements=ele_ref2)

    line_ref1.append_element(element=xt.Marker(), name='endmarker')
    line_ref2.append_element(element=xt.Marker(), name='endmarker')

    line_ref1.configure_bend_model(edge=model)
    line_ref2.configure_bend_model(edge=model)

    line_ref1.build_tracker()
    line_ref2.build_tracker()

    line_ref1.track(p1, ele_start=0, ele_stop='endmarker')
    line_ref2.track(p1_ref2, ele_start=0, ele_stop='endmarker')

    xo.assert_allclose(p1.x, p1_ref2.x, atol=0, rtol=1e-5)
    xo.assert_allclose(p1.y, p1_ref2.y, atol=0, rtol=1e-1)          #?????????????
    xo.assert_allclose(p1_ref2.mass, p1.mass, atol=0, rtol=1e-14)
    xo.assert_allclose(p1_ref2.charge, p1.charge, atol=0, rtol=1e-14)
    xo.assert_allclose(p1_ref2.energy, p1.energy, atol=0, rtol=1e-14)
    xo.assert_allclose(p1_ref2.rvv * p1_ref2.beta0, p1.rvv * p1.beta0, atol=0, rtol=1e-14)

    # Check absolute time of arrival
    t0_ref1 = p1.s / (p1.beta0 * clight)           # Absolute reference time of arrival
    t0_ref2 = p1_ref2.s / (p1_ref2.beta0 * clight) # Absolute reference time of arrival
    dt_ref1 = -p1.zeta / (p1.beta0 * clight)           # Arrival time relative to reference
    dt_ref2 = -p1_ref2.zeta / (p1_ref2.beta0 * clight) # Arrival time relative to reference

    xo.assert_allclose(t0_ref1 + dt_ref1, t0_ref2 + dt_ref2, atol=1e-11, rtol=0)


def test_multispecies_quadrupole():

    p_ref1 = xt.Particles(
        mass0=xt.PROTON_MASS_EV,
        q0=2,
        gamma0=1.2)

    p_ref2 = xt.Particles(
        mass0=2*xt.PROTON_MASS_EV,
        q0=3,
        gamma0=1.5)

    # Build a particle referred to reference 1
    p1 = p_ref1.copy()
    p1.x = 1e-3
    p1.y = 2e-3
    p1.delta = 0.5
    P_p1 = (1 + p1.delta) * p1.p0c

    # Build the same particle referred to reference 2
    p1_ref2 = p_ref2.copy()
    p1_ref2_mass_ratio = p1.mass0 / p1_ref2.mass0
    p1_ref2_charge_ratio = p1.q0 / p1_ref2.q0

    p1_ref2.x = p1.x
    p1_ref2.y = p1.y
    p1_ref2.zeta = p1.zeta / p1.beta0 * p1_ref2.beta0
    p1_ref2.charge_ratio = p1_ref2_charge_ratio
    p1_ref2.chi = p1_ref2_charge_ratio / p1_ref2_mass_ratio
    p1_ref2.delta = P_p1 / p1_ref2_mass_ratio / p1_ref2.p0c - 1

    xo.assert_allclose(p1_ref2.mass, p1.mass, atol=0, rtol=1e-14)
    xo.assert_allclose(p1_ref2.charge, p1.charge, atol=0, rtol=1e-14)
    xo.assert_allclose(p1_ref2.energy, p1.energy, atol=0, rtol=1e-14)
    xo.assert_allclose(p1_ref2.rvv * p1_ref2.beta0, p1.rvv * p1.beta0, atol=0, rtol=1e-14)

    xo.assert_allclose(p1.rpp, 1 / (1 + p1.delta), atol=0, rtol=1e-14)
    xo.assert_allclose(p1_ref2.rpp, 1 / (1 + p1_ref2.delta), atol=0, rtol=1e-14)

    p1c = p1.p0c / p1.rpp * p1.mass_ratio
    p1c_ref2 = p1_ref2.p0c / p1_ref2.rpp * p1_ref2.mass_ratio
    xo.assert_allclose(p1c, p1c_ref2, atol=0, rtol=1e-14)

    model = 'bend-kick-bend'
    model = 'rot-kick-rot'
    model = 'expanded'
    L_bend = 1.
    G_Tm = 0.1
    Gs_Tm = -0.05

    P0_J_ref1 = p_ref1.p0c[0] / clight * qe
    k1l_ref1 = G_Tm * qe * p_ref1.charge[0] / P0_J_ref1
    k1sl_ref1 = Gs_Tm * qe * p_ref1.charge[0] / P0_J_ref1


    P0_J_ref2 = p_ref2.p0c[0] / clight * qe
    k1l_ref2 = G_Tm * qe * p_ref2.charge[0] / P0_J_ref2
    k1sl_ref2 = Gs_Tm * qe * p_ref2.charge[0] / P0_J_ref2

    n_slices = 10

    quad_ref1 = xt.Quadrupole(k1=k1l_ref1, k1s=k1sl_ref1, length=L_bend/n_slices)
    quad_ref2 = xt.Quadrupole(k1=k1l_ref2, k1s=k1sl_ref2, length=L_bend/n_slices)

    ele_ref1 = []
    for ii in range(n_slices):
        ele_ref1.append(xt.Drift(length=L_bend/n_slices/2))
        ele_ref1.append(quad_ref1)
        ele_ref1.append(xt.Drift(length=L_bend/n_slices/2))

    ele_ref2 = []
    for ii in range(n_slices):
        ele_ref2.append(xt.Drift(length=L_bend/n_slices/2))
        ele_ref2.append(quad_ref2)
        ele_ref2.append(xt.Drift(length=L_bend/n_slices/2))

    line_ref1 = xt.Line(elements=ele_ref1)
    line_ref2 = xt.Line(elements=ele_ref2)

    line_ref1.append_element(element=xt.Marker(), name='endmarker')
    line_ref2.append_element(element=xt.Marker(), name='endmarker')

    line_ref1.configure_bend_model(core=model)
    line_ref2.configure_bend_model(core=model)

    # line_ref1.config.XTRACK_USE_EXACT_DRIFTS = True
    # line_ref2.config.XTRACK_USE_EXACT_DRIFTS = True

    line_ref1.build_tracker()
    line_ref2.build_tracker()

    line_ref1.track(p1, ele_start=0, ele_stop='endmarker')
    line_ref2.track(p1_ref2, ele_start=0, ele_stop='endmarker')

    xo.assert_allclose(p1.x, p1_ref2.x, atol=0, rtol=1e-10)
    xo.assert_allclose(p1.y, p1_ref2.y, atol=0, rtol=1e-10)
    xo.assert_allclose(p1_ref2.mass, p1.mass, atol=0, rtol=1e-14)
    xo.assert_allclose(p1_ref2.charge, p1.charge, atol=0, rtol=1e-14)
    xo.assert_allclose(p1_ref2.energy, p1.energy, atol=0, rtol=1e-14)
    xo.assert_allclose(p1_ref2.rvv * p1_ref2.beta0, p1.rvv * p1.beta0, atol=0, rtol=1e-14)

    # Check absolute time of arrival
    t0_ref1 = p1.s / (p1.beta0 * clight)           # Absolute reference time of arrival
    t0_ref2 = p1_ref2.s / (p1_ref2.beta0 * clight) # Absolute reference time of arrival
    dt_ref1 = -p1.zeta / (p1.beta0 * clight)           # Arrival time relative to reference
    dt_ref2 = -p1_ref2.zeta / (p1_ref2.beta0 * clight) # Arrival time relative to reference

    xo.assert_allclose(t0_ref1 + dt_ref1, t0_ref2 + dt_ref2, atol=1e-11, rtol=0)

def test_multispecies_sextupole():

    p_ref1 = xt.Particles(
        mass0=xt.PROTON_MASS_EV,
        q0=2,
        gamma0=1.2)

    p_ref2 = xt.Particles(
        mass0=2*xt.PROTON_MASS_EV,
        q0=3,
        gamma0=1.5)

    # Build a particle referred to reference 1
    p1 = p_ref1.copy()
    p1.x = 1e-3
    p1.y = 2e-3
    p1.delta = 0.5
    P_p1 = (1 + p1.delta) * p1.p0c

    # Build the same particle referred to reference 2
    p1_ref2 = p_ref2.copy()
    p1_ref2_mass_ratio = p1.mass0 / p1_ref2.mass0
    p1_ref2_charge_ratio = p1.q0 / p1_ref2.q0

    p1_ref2.x = p1.x
    p1_ref2.y = p1.y
    p1_ref2.zeta = p1.zeta / p1.beta0 * p1_ref2.beta0
    p1_ref2.charge_ratio = p1_ref2_charge_ratio
    p1_ref2.chi = p1_ref2_charge_ratio / p1_ref2_mass_ratio
    p1_ref2.delta = P_p1 / p1_ref2_mass_ratio / p1_ref2.p0c - 1

    xo.assert_allclose(p1_ref2.mass, p1.mass, atol=0, rtol=1e-14)
    xo.assert_allclose(p1_ref2.charge, p1.charge, atol=0, rtol=1e-14)
    xo.assert_allclose(p1_ref2.energy, p1.energy, atol=0, rtol=1e-14)
    xo.assert_allclose(p1_ref2.rvv * p1_ref2.beta0, p1.rvv * p1.beta0, atol=0, rtol=1e-14)

    xo.assert_allclose(p1.rpp, 1 / (1 + p1.delta), atol=0, rtol=1e-14)
    xo.assert_allclose(p1_ref2.rpp, 1 / (1 + p1_ref2.delta), atol=0, rtol=1e-14)

    p1c = p1.p0c / p1.rpp * p1.mass_ratio
    p1c_ref2 = p1_ref2.p0c / p1_ref2.rpp * p1_ref2.mass_ratio
    xo.assert_allclose(p1c, p1c_ref2, atol=0, rtol=1e-14)

    model = 'bend-kick-bend'
    model = 'rot-kick-rot'
    model = 'expanded'
    L_bend = 1.
    S_Tm2 = 0.01
    Ss_Tm2 = -0.05

    P0_J_ref1 = p_ref1.p0c[0] / clight * qe
    k2_ref1 = S_Tm2 * qe * p_ref1.charge[0] / P0_J_ref1
    k2s_ref1 = Ss_Tm2 * qe * p_ref1.charge[0] / P0_J_ref1

    P0_J_ref2 = p_ref2.p0c[0] / clight * qe
    k2_ref2 = S_Tm2 * qe * p_ref2.charge[0] / P0_J_ref2
    k2s_ref2 = Ss_Tm2 * qe * p_ref2.charge[0] / P0_J_ref2

    n_slices = 10

    sext_ref1 = xt.Sextupole(k2=k2_ref1, k2s=k2s_ref1, length=L_bend/n_slices)
    sext_ref2 = xt.Sextupole(k2=k2_ref2, k2s=k2s_ref2, length=L_bend/n_slices)

    ele_ref1 = []
    for ii in range(n_slices):
        ele_ref1.append(xt.Drift(length=L_bend/n_slices/2))
        ele_ref1.append(sext_ref1)
        ele_ref1.append(xt.Drift(length=L_bend/n_slices/2))

    ele_ref2 = []
    for ii in range(n_slices):
        ele_ref2.append(xt.Drift(length=L_bend/n_slices/2))
        ele_ref2.append(sext_ref2)
        ele_ref2.append(xt.Drift(length=L_bend/n_slices/2))

    line_ref1 = xt.Line(elements=ele_ref1)
    line_ref2 = xt.Line(elements=ele_ref2)

    line_ref1.append_element(element=xt.Marker(), name='endmarker')
    line_ref2.append_element(element=xt.Marker(), name='endmarker')

    line_ref1.configure_bend_model(core=model)
    line_ref2.configure_bend_model(core=model)

    # line_ref1.config.XTRACK_USE_EXACT_DRIFTS = True
    # line_ref2.config.XTRACK_USE_EXACT_DRIFTS = True

    line_ref1.build_tracker()
    line_ref2.build_tracker()

    line_ref1.track(p1, ele_start=0, ele_stop='endmarker')
    line_ref2.track(p1_ref2, ele_start=0, ele_stop='endmarker')

    xo.assert_allclose(p1.x, p1_ref2.x, atol=0, rtol=1e-10)
    xo.assert_allclose(p1.y, p1_ref2.y, atol=0, rtol=1e-10)
    xo.assert_allclose(p1_ref2.mass, p1.mass, atol=0, rtol=1e-14)
    xo.assert_allclose(p1_ref2.charge, p1.charge, atol=0, rtol=1e-14)
    xo.assert_allclose(p1_ref2.energy, p1.energy, atol=0, rtol=1e-14)
    xo.assert_allclose(p1_ref2.rvv * p1_ref2.beta0, p1.rvv * p1.beta0, atol=0, rtol=1e-14)

    # Check absolute time of arrival
    t0_ref1 = p1.s / (p1.beta0 * clight)           # Absolute reference time of arrival
    t0_ref2 = p1_ref2.s / (p1_ref2.beta0 * clight) # Absolute reference time of arrival
    dt_ref1 = -p1.zeta / (p1.beta0 * clight)           # Arrival time relative to reference
    dt_ref2 = -p1_ref2.zeta / (p1_ref2.beta0 * clight) # Arrival time relative to reference

    xo.assert_allclose(t0_ref1 + dt_ref1, t0_ref2 + dt_ref2, atol=1e-11, rtol=0)

def test_multispecies_octupole():

    p_ref1 = xt.Particles(
        mass0=xt.PROTON_MASS_EV,
        q0=2,
        gamma0=1.2)

    p_ref2 = xt.Particles(
        mass0=2*xt.PROTON_MASS_EV,
        q0=3,
        gamma0=1.5)

    # Build a particle referred to reference 1
    p1 = p_ref1.copy()
    p1.x = 1e-2
    p1.y = 2e-2
    p1.delta = 0.5
    P_p1 = (1 + p1.delta) * p1.p0c

    # Build the same particle referred to reference 2
    p1_ref2 = p_ref2.copy()
    p1_ref2_mass_ratio = p1.mass0 / p1_ref2.mass0
    p1_ref2_charge_ratio = p1.q0 / p1_ref2.q0

    p1_ref2.x = p1.x
    p1_ref2.y = p1.y
    p1_ref2.zeta = p1.zeta / p1.beta0 * p1_ref2.beta0
    p1_ref2.charge_ratio = p1_ref2_charge_ratio
    p1_ref2.chi = p1_ref2_charge_ratio / p1_ref2_mass_ratio
    p1_ref2.delta = P_p1 / p1_ref2_mass_ratio / p1_ref2.p0c - 1

    xo.assert_allclose(p1_ref2.mass, p1.mass, atol=0, rtol=1e-14)
    xo.assert_allclose(p1_ref2.charge, p1.charge, atol=0, rtol=1e-14)
    xo.assert_allclose(p1_ref2.energy, p1.energy, atol=0, rtol=1e-14)
    xo.assert_allclose(p1_ref2.rvv * p1_ref2.beta0, p1.rvv * p1.beta0, atol=0, rtol=1e-14)

    xo.assert_allclose(p1.rpp, 1 / (1 + p1.delta), atol=0, rtol=1e-14)
    xo.assert_allclose(p1_ref2.rpp, 1 / (1 + p1_ref2.delta), atol=0, rtol=1e-14)

    p1c = p1.p0c / p1.rpp * p1.mass_ratio
    p1c_ref2 = p1_ref2.p0c / p1_ref2.rpp * p1_ref2.mass_ratio
    xo.assert_allclose(p1c, p1c_ref2, atol=0, rtol=1e-14)

    L_bend = 1.
    O_Tm3 = 1
    Os_Tm3 = -2

    P0_J_ref1 = p_ref1.p0c[0] / clight * qe
    k3_ref1 = O_Tm3 * qe * p_ref1.charge[0] / P0_J_ref1
    k3s_ref1 = Os_Tm3 * qe * p_ref1.charge[0] / P0_J_ref1

    P0_J_ref2 = p_ref2.p0c[0] / clight * qe
    k3_ref2 = O_Tm3 * qe * p_ref2.charge[0] / P0_J_ref2
    k3s_ref2 = Os_Tm3 * qe * p_ref2.charge[0] / P0_J_ref2

    n_slices = 10

    oct_ref1 = xt.Octupole(k3=k3_ref1, k3s=k3s_ref1, length=L_bend/n_slices)
    oct_ref2 = xt.Octupole(k3=k3_ref2, k3s=k3s_ref2, length=L_bend/n_slices)

    ele_ref1 = []
    for ii in range(n_slices):
        ele_ref1.append(xt.Drift(length=L_bend/n_slices/2))
        ele_ref1.append(oct_ref1)
        ele_ref1.append(xt.Drift(length=L_bend/n_slices/2))

    ele_ref2 = []
    for ii in range(n_slices):
        ele_ref2.append(xt.Drift(length=L_bend/n_slices/2))
        ele_ref2.append(oct_ref2)
        ele_ref2.append(xt.Drift(length=L_bend/n_slices/2))

    line_ref1 = xt.Line(elements=ele_ref1)
    line_ref2 = xt.Line(elements=ele_ref2)

    line_ref1.append_element(element=xt.Marker(), name='endmarker')
    line_ref2.append_element(element=xt.Marker(), name='endmarker')

    line_ref1.config.XTRACK_USE_EXACT_DRIFTS = True
    line_ref2.config.XTRACK_USE_EXACT_DRIFTS = True

    line_ref1.build_tracker()
    line_ref2.build_tracker()

    line_ref1.track(p1, ele_start=0, ele_stop='endmarker')
    line_ref2.track(p1_ref2, ele_start=0, ele_stop='endmarker')

    xo.assert_allclose(p1.x, p1_ref2.x, atol=0, rtol=1e-12)
    xo.assert_allclose(p1.y, p1_ref2.y, atol=0, rtol=1e-12)
    xo.assert_allclose(p1_ref2.mass, p1.mass, atol=0, rtol=1e-14)
    xo.assert_allclose(p1_ref2.charge, p1.charge, atol=0, rtol=1e-14)
    xo.assert_allclose(p1_ref2.energy, p1.energy, atol=0, rtol=1e-14)
    xo.assert_allclose(p1_ref2.rvv * p1_ref2.beta0, p1.rvv * p1.beta0, atol=0, rtol=1e-14)

    # Check absolute time of arrival
    t0_ref1 = p1.s / (p1.beta0 * clight)           # Absolute reference time of arrival
    t0_ref2 = p1_ref2.s / (p1_ref2.beta0 * clight) # Absolute reference time of arrival
    dt_ref1 = -p1.zeta / (p1.beta0 * clight)           # Arrival time relative to reference
    dt_ref2 = -p1_ref2.zeta / (p1_ref2.beta0 * clight) # Arrival time relative to reference

    xo.assert_allclose(t0_ref1 + dt_ref1, t0_ref2 + dt_ref2, atol=1e-11, rtol=0)

def test_multispecies_cavity():

    p_ref1 = xt.Particles(
        mass0=xt.PROTON_MASS_EV,
        q0=2,
        gamma0=1.2)

    p_ref2 = xt.Particles(
        mass0=2*xt.PROTON_MASS_EV,
        q0=3,
        gamma0=1.5)

    # Build a particle referred to reference 1
    p1 = p_ref1.copy()
    p1.x = 1e-2
    p1.y = 2e-2
    p1.delta = 0.5
    P_p1 = (1 + p1.delta) * p1.p0c

    # Build the same particle referred to reference 2
    p1_ref2 = p_ref2.copy()
    p1_ref2_mass_ratio = p1.mass0 / p1_ref2.mass0
    p1_ref2_charge_ratio = p1.q0 / p1_ref2.q0

    p1_ref2.x = p1.x
    p1_ref2.y = p1.y
    p1_ref2.zeta = p1.zeta / p1.beta0 * p1_ref2.beta0
    p1_ref2.charge_ratio = p1_ref2_charge_ratio
    p1_ref2.chi = p1_ref2_charge_ratio / p1_ref2_mass_ratio
    p1_ref2.delta = P_p1 / p1_ref2_mass_ratio / p1_ref2.p0c - 1

    xo.assert_allclose(p1_ref2.mass, p1.mass, atol=0, rtol=1e-14)
    xo.assert_allclose(p1_ref2.charge, p1.charge, atol=0, rtol=1e-14)
    xo.assert_allclose(p1_ref2.energy, p1.energy, atol=0, rtol=1e-14)
    xo.assert_allclose(p1_ref2.rvv * p1_ref2.beta0, p1.rvv * p1.beta0, atol=0, rtol=1e-14)

    xo.assert_allclose(p1.rpp, 1 / (1 + p1.delta), atol=0, rtol=1e-14)
    xo.assert_allclose(p1_ref2.rpp, 1 / (1 + p1_ref2.delta), atol=0, rtol=1e-14)

    p1c = p1.p0c / p1.rpp * p1.mass_ratio
    p1c_ref2 = p1_ref2.p0c / p1_ref2.rpp * p1_ref2.mass_ratio
    xo.assert_allclose(p1c, p1c_ref2, atol=0, rtol=1e-14)

    s_cav = 100
    frequency = 400e6


    lag2 = 180 / np.pi *  2 * np.pi * frequency * s_cav / clight * (1/p1_ref2.beta0 - 1/p1.beta0)


    cav_ref1 = xt.Cavity(voltage=1e6, frequency=frequency)
    cav_ref2 = xt.Cavity(voltage=1e6, frequency=frequency, lag=lag2)

    ele_ref1 = [xt.Drift(length=s_cav), cav_ref1, xt.Drift(length=0)]
    ele_ref2 = [xt.Drift(length=s_cav), cav_ref2, xt.Drift(length=0)]

    line_ref1 = xt.Line(elements=ele_ref1)
    line_ref2 = xt.Line(elements=ele_ref2)

    line_ref1.append_element(element=xt.Marker(), name='endmarker')
    line_ref2.append_element(element=xt.Marker(), name='endmarker')

    line_ref1.build_tracker()
    line_ref2.build_tracker()

    line_ref1.track(p1, ele_start=0, ele_stop='endmarker')
    line_ref2.track(p1_ref2, ele_start=0, ele_stop='endmarker')



    # Check absolute time of arrival
    t0_ref1 = p1.s / (p1.beta0 * clight)           # Absolute reference time of arrival
    t0_ref2 = p1_ref2.s / (p1_ref2.beta0 * clight) # Absolute reference time of arrival
    dt_ref1 = -p1.zeta / (p1.beta0 * clight)           # Arrival time relative to reference
    dt_ref2 = -p1_ref2.zeta / (p1_ref2.beta0 * clight) # Arrival time relative to reference

    xo.assert_allclose(t0_ref1 + dt_ref1, t0_ref2 + dt_ref2, atol=1e-11, rtol=0)
    xo.assert_allclose(p1_ref2.zeta,
        (1 - p1_ref2.beta0 / p1.beta0) * p1.s + p1_ref2.beta0 / p1.beta0 * p1.zeta,
        atol=1e-11, rtol=0)

    xo.assert_allclose(p1.x, p1_ref2.x, atol=0, rtol=1e-12)
    xo.assert_allclose(p1.y, p1_ref2.y, atol=0, rtol=1e-12)
    xo.assert_allclose(p1_ref2.mass, p1.mass, atol=0, rtol=1e-14)
    xo.assert_allclose(p1_ref2.charge, p1.charge, atol=0, rtol=1e-14)
    xo.assert_allclose(p1_ref2.energy, p1.energy, atol=0, rtol=1e-14)
    xo.assert_allclose(p1_ref2.rvv * p1_ref2.beta0, p1.rvv * p1.beta0, atol=0, rtol=1e-14)


import xtrack as xt
import xobjects as xo

from xtrack.beam_elements.magnets import Magnet

mm = Magnet(length=2.0, k0=0.0, k1=0.0, h=0.0)
mm.integrator = 'teapot'

p0 = xt.Particles(kinetic_energy0=50e6,
                  x=[1e-3, -1e-3], y=2e-3, zeta=1e-2, px=10e-3, py=20e-3, delta=1e-2)

mm.compile_kernels()

# Expanded drift
mm.model = 'drift-kick-drift-expanded'
p_test = p0.copy()
p_ref = p0.copy()

eref = xt.Drift(length=2.0)
mm.track(p_test)
eref.track(p_ref)

xo.assert_allclose(p_test.s, 2.0, atol=0, rtol=1e-7)
xo.assert_allclose(p_ref.s, 2.0, atol=0, rtol=1e-7)
xo.assert_allclose(p_test.x, p_ref.x, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.y, p_ref.y, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.zeta, p_ref.zeta, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.px, p_ref.px, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.py, p_ref.py, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.delta, p_ref.delta, atol=1e-15, rtol=0)

# Exact drift
mm.model = 'drift-kick-drift-exact'
p_test = p0.copy()
p_ref = p0.copy()

eref = xt.Solenoid(length=2.0) # Solenoid is exact drift when off
mm.track(p_test)
eref.track(p_ref)

xo.assert_allclose(p_test.s, 2.0, atol=0, rtol=1e-7)
xo.assert_allclose(p_ref.s, 2.0, atol=0, rtol=1e-7)
xo.assert_allclose(p_test.x, p_ref.x, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.y, p_ref.y, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.zeta, p_ref.zeta, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.px, p_ref.px, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.py, p_ref.py, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.delta, p_ref.delta, atol=1e-15, rtol=0)

# Test backtracking
line = xt.Line(elements=[mm])
line.track(p_test, backtrack=True)
xo.assert_allclose(p_test.s, 0.0, atol=1e-7, rtol=0)
xo.assert_allclose(p_test.x, p0.x, atol=5e-14, rtol=0)
xo.assert_allclose(p_test.y, p0.y, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.zeta, p0.zeta, atol=1e-14, rtol=0)
xo.assert_allclose(p_test.px, p0.px, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.py, p0.py, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.delta, p0.delta, atol=1e-15, rtol=0)

# Sextupole
mm.model = 'drift-kick-drift-expanded'
mm.k2 = 3.
mm.k2s = 5.

p_test = p0.copy()
p_ref = p0.copy()

eref = xt.Sextupole(length=2.0, k2=3., k2s=5.)
mm.track(p_test)
eref.track(p_ref)

xo.assert_allclose(p_test.s, 2.0, atol=0, rtol=1e-7)
xo.assert_allclose(p_ref.s, 2.0, atol=0, rtol=1e-7)
xo.assert_allclose(p_test.x, p_ref.x, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.y, p_ref.y, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.zeta, p_ref.zeta, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.px, p_ref.px, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.py, p_ref.py, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.delta, p_ref.delta, atol=1e-15, rtol=0)

# Test backtracking
line = xt.Line(elements=[mm])
line.track(p_test, backtrack=True)
xo.assert_allclose(p_test.s, 0.0, atol=1e-7, rtol=0)
xo.assert_allclose(p_test.x, p0.x, atol=5e-14, rtol=0)
xo.assert_allclose(p_test.y, p0.y, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.zeta, p0.zeta, atol=1e-14, rtol=0)
xo.assert_allclose(p_test.px, p0.px, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.py, p0.py, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.delta, p0.delta, atol=1e-15, rtol=0)

# Sextupole more kicks
mm.num_multipole_kicks = 5
eref.num_multipole_kicks = 5
mm.edge_entry_active = False
mm.edge_exit_active = False
mm.integrator = 'teapot'
eref.integrator = 'teapot'

p_test = p0.copy()
p_ref = p0.copy()

mm.track(p_test)
eref.track(p_ref)

xo.assert_allclose(p_test.s, 2.0, atol=0, rtol=1e-7)
xo.assert_allclose(p_ref.s, 2.0, atol=0, rtol=1e-7)
xo.assert_allclose(p_test.x, p_ref.x, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.y, p_ref.y, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.zeta, p_ref.zeta, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.px, p_ref.px, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.py, p_ref.py, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.delta, p_ref.delta, atol=1e-15, rtol=0)

# Test backtracking
line = xt.Line(elements=[mm])
line.track(p_test, backtrack=True)
xo.assert_allclose(p_test.s, 0.0, atol=1e-7, rtol=0)
xo.assert_allclose(p_test.x, p0.x, atol=5e-14, rtol=0)
xo.assert_allclose(p_test.y, p0.y, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.zeta, p0.zeta, atol=1e-14, rtol=0)
xo.assert_allclose(p_test.px, p0.px, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.py, p0.py, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.delta, p0.delta, atol=1e-15, rtol=0)

# Sextupole done with knl
mm.k2 = 0.
mm.knl = [0., 0., 3.*2., 0., 0., 0.]
mm.num_multipole_kicks = 5
eref.num_multipole_kicks = 5

p_test = p0.copy()
p_ref = p0.copy()

mm.track(p_test)
eref.track(p_ref)

xo.assert_allclose(p_test.s, 2.0, atol=0, rtol=1e-7)
xo.assert_allclose(p_ref.s, 2.0, atol=0, rtol=1e-7)
xo.assert_allclose(p_test.x, p_ref.x, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.y, p_ref.y, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.zeta, p_ref.zeta, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.px, p_ref.px, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.py, p_ref.py, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.delta, p_ref.delta, atol=1e-15, rtol=0)

# Test backtracking
line = xt.Line(elements=[mm])
line.track(p_test, backtrack=True)
xo.assert_allclose(p_test.s, 0.0, atol=1e-7, rtol=0)
xo.assert_allclose(p_test.x, p0.x, atol=5e-14, rtol=0)
xo.assert_allclose(p_test.y, p0.y, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.zeta, p0.zeta, atol=1e-14, rtol=0)
xo.assert_allclose(p_test.px, p0.px, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.py, p0.py, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.delta, p0.delta, atol=1e-15, rtol=0)


# Add skew sextupole component in ksl
mm.ksl[2] = -2.
eref.ksl[2] = -2.

p_test = p0.copy()
p_ref = p0.copy()

mm.track(p_test)
eref.track(p_ref)

xo.assert_allclose(p_test.s, 2.0, atol=0, rtol=1e-7)
xo.assert_allclose(p_ref.s, 2.0, atol=0, rtol=1e-7)
xo.assert_allclose(p_test.x, p_ref.x, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.y, p_ref.y, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.zeta, p_ref.zeta, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.px, p_ref.px, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.py, p_ref.py, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.delta, p_ref.delta, atol=1e-15, rtol=0)

# Test backtracking
line = xt.Line(elements=[mm])
line.track(p_test, backtrack=True)
xo.assert_allclose(p_test.s, 0.0, atol=1e-7, rtol=0)
xo.assert_allclose(p_test.x, p0.x, atol=5e-14, rtol=0)
xo.assert_allclose(p_test.y, p0.y, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.zeta, p0.zeta, atol=1e-14, rtol=0)
xo.assert_allclose(p_test.px, p0.px, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.py, p0.py, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.delta, p0.delta, atol=1e-15, rtol=0)

# Quadrupole
mm.model = 'mat-kick-mat'
mm.k0 = 0
mm.k1 = 3.
mm.k2 = 0.
mm.k2s = 0.
mm.knl = 0.
mm.ksl = 0.
mm.num_multipole_kicks = 1

eref = xt.Quadrupole(length=2.0, k1=3.)
eref.num_multipole_kicks = 1

p_test = p0.copy()
p_ref = p0.copy()

mm.track(p_test)
eref.track(p_ref)

xo.assert_allclose(p_test.s, 2.0, atol=0, rtol=1e-7)
xo.assert_allclose(p_ref.s, 2.0, atol=0, rtol=1e-7)
xo.assert_allclose(p_test.x, p_ref.x, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.y, p_ref.y, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.zeta, p_ref.zeta, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.px, p_ref.px, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.py, p_ref.py, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.delta, p_ref.delta, atol=1e-15, rtol=0)

# Test backtracking
line = xt.Line(elements=[mm])
line.track(p_test, backtrack=True)
xo.assert_allclose(p_test.s, 0.0, atol=1e-7, rtol=0)
xo.assert_allclose(p_test.x, p0.x, atol=5e-14, rtol=0)
xo.assert_allclose(p_test.y, p0.y, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.zeta, p0.zeta, atol=1e-14, rtol=0)
xo.assert_allclose(p_test.px, p0.px, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.py, p0.py, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.delta, p0.delta, atol=1e-15, rtol=0)

# Bend
mm.model = 'rot-kick-rot'
mm.integrator = 'yoshida4'
mm.num_multipole_kicks = 15
mm.length=2.0
mm.h = 0.05
mm.k1 = -0.3

eref = xt.Bend(length=2.0, k1=-0.3, h=0.05)
eref.num_multipole_kicks = 15

p_test = p0.copy()
p_ref = p0.copy()

mm.track(p_test)
eref.track(p_ref)

xo.assert_allclose(p_test.s, 2.0, atol=0, rtol=1e-7)
xo.assert_allclose(p_ref.s, 2.0, atol=0, rtol=1e-7)
xo.assert_allclose(p_test.x, p_ref.x, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.y, p_ref.y, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.zeta, p_ref.zeta, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.px, p_ref.px, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.py, p_ref.py, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.delta, p_ref.delta, atol=1e-15, rtol=0)

# Test backtracking
line = xt.Line(elements=[mm])
line.track(p_test, backtrack=True)
xo.assert_allclose(p_test.s, 0.0, atol=1e-7, rtol=0)
xo.assert_allclose(p_test.x, p0.x, atol=5e-14, rtol=0)
xo.assert_allclose(p_test.y, p0.y, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.zeta, p0.zeta, atol=1e-14, rtol=0)
xo.assert_allclose(p_test.px, p0.px, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.py, p0.py, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.delta, p0.delta, atol=1e-15, rtol=0)

# Bend auto no kicks
mm.model = 'bend-kick-bend'
mm.integrator = 'yoshida4'
mm.num_multipole_kicks = 0
mm.h = 0.05
mm.k1 = 0

eref = xt.Bend(length=2.0, h=0.05)
eref.num_multipole_kicks = 0

p_test = p0.copy()
p_ref = p0.copy()

mm.track(p_test)
eref.track(p_ref)

xo.assert_allclose(p_test.s, 2.0, atol=0, rtol=1e-7)
xo.assert_allclose(p_ref.s, 2.0, atol=0, rtol=1e-7)
xo.assert_allclose(p_test.x, p_ref.x, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.y, p_ref.y, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.zeta, p_ref.zeta, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.px, p_ref.px, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.py, p_ref.py, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.delta, p_ref.delta, atol=1e-15, rtol=0)

# Test backtracking
line = xt.Line(elements=[mm])
line.track(p_test, backtrack=True)
xo.assert_allclose(p_test.s, 0.0, atol=1e-7, rtol=0)
xo.assert_allclose(p_test.x, p0.x, atol=5e-14, rtol=0)
xo.assert_allclose(p_test.y, p0.y, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.zeta, p0.zeta, atol=1e-14, rtol=0)
xo.assert_allclose(p_test.px, p0.px, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.py, p0.py, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.delta, p0.delta, atol=1e-15, rtol=0)

# Bend auto quad kick
mm.model = 'bend-kick-bend'
mm.integrator = 'yoshida4'
mm.num_multipole_kicks = 1
mm.h = 0.05
mm.k1 = 0.3

eref = xt.Bend(length=2.0, h=0.05, k1=0.3)
eref.num_multipole_kicks = 1
eref.edge_entry_active = False
eref.edge_exit_active = False
eref.model = 'bend-kick-bend'

p_test = p0.copy()
p_ref = p0.copy()

mm.track(p_test)
eref.track(p_ref)

xo.assert_allclose(p_test.s, 2.0, atol=0, rtol=1e-7)
xo.assert_allclose(p_ref.s, 2.0, atol=0, rtol=1e-7)
xo.assert_allclose(p_test.x, p_ref.x, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.y, p_ref.y, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.zeta, p_ref.zeta, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.px, p_ref.px, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.py, p_ref.py, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.delta, p_ref.delta, atol=1e-15, rtol=0)

# Test backtracking
line = xt.Line(elements=[mm])
line.track(p_test, backtrack=True)
xo.assert_allclose(p_test.s, 0.0, atol=1e-7, rtol=0)
xo.assert_allclose(p_test.x, p0.x, atol=5e-14, rtol=0)
xo.assert_allclose(p_test.y, p0.y, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.zeta, p0.zeta, atol=1e-14, rtol=0)
xo.assert_allclose(p_test.px, p0.px, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.py, p0.py, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.delta, p0.delta, atol=1e-15, rtol=0)

# Bend dip quad kick
mm.integrator = 'yoshida4'
mm.num_multipole_kicks = 10
mm.h = 0.1
mm.k1 = 0.3
mm.k0 = 0.2

eref = xt.Bend(length=2.0, h=0.1, k1=0.3, k0=0.2)
eref.num_multipole_kicks = 10
eref.edge_entry_active = False
eref.edge_exit_active = False

for model in ['bend-kick-bend', 'rot-kick-rot', 'expanded']:
    mm.model = model
    eref.model = model

    p_test = p0.copy()
    p_ref = p0.copy()

    mm.track(p_test)
    eref.track(p_ref)

    xo.assert_allclose(p_test.s, 2.0, atol=0, rtol=1e-7)
    xo.assert_allclose(p_ref.s, 2.0, atol=0, rtol=1e-7)
    xo.assert_allclose(p_test.x, p_ref.x, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test.y, p_ref.y, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test.zeta, p_ref.zeta, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test.px, p_ref.px, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test.py, p_ref.py, atol=1e-15, rtol=0)
    xo.assert_allclose(p_test.delta, p_ref.delta, atol=1e-15, rtol=0)

    # Test backtracking
    line = xt.Line(elements=[mm])
    line.track(p_test, backtrack=True)
    xo.assert_allclose(p_test.s, 0.0, rtol=0, atol=1e-7)
    xo.assert_allclose(p_test.x, p0.x, atol=5e-14, rtol=0)
    xo.assert_allclose(p_test.y, p0.y, atol=5e-14, rtol=0)
    xo.assert_allclose(p_test.zeta, p0.zeta, atol=5e-14, rtol=0)
    xo.assert_allclose(p_test.px, p0.px, atol=5e-14, rtol=0)
    xo.assert_allclose(p_test.py, p0.py, atol=5e-14, rtol=0)
    xo.assert_allclose(p_test.delta, p0.delta, atol=1e-15, rtol=0)

# Bend dip quad kick with multipoles
mm.integrator = 'yoshida4'
mm.num_multipole_kicks = 10
mm.h = 0.1
mm.k1 = 0.3
mm.k0 = 0.2

mm.k2 = 0.1
mm.k3 = 0.15

mm.k0s = 0.02
mm.k1s = 0.03
mm.k2s = 0.01
mm.k3s = 0.02

mm.knl = [0.1, 0.2, 0.3, 0.4, 0.5, 0.6]
mm.ksl = [0.6, 0.5, 0.4, 0.15, 0.2, 0.1]

eref = xt.Bend(length=2.0, h=0.1, k1=0.3, k0=0.2)
eref.num_multipole_kicks = 10
eref.edge_entry_active = False
eref.edge_exit_active = False
eref.knl = [0.1, 0.2, 0.3 + 0.1*2, 0.4 + 0.15*2, 0.5, 0.6]
eref.ksl = [0.6 + 0.02*2, 0.5 + 0.03*2, 0.4 + 0.01*2, 0.15 + 0.02*2, 0.2, 0.1]

for model in ['bend-kick-bend', 'rot-kick-rot']:
    mm.model = model
    eref.model = model

    p_test = p0.copy()
    p_ref = p0.copy()

    mm.track(p_test)
    eref.track(p_ref)

    xo.assert_allclose(p_test.s, 2.0, atol=0, rtol=1e-7)
    xo.assert_allclose(p_ref.s, 2.0, atol=0, rtol=1e-7)
    xo.assert_allclose(p_test.x, p_ref.x, atol=1e-14, rtol=0)
    xo.assert_allclose(p_test.y, p_ref.y, atol=1e-14, rtol=0)
    xo.assert_allclose(p_test.zeta, p_ref.zeta, atol=1e-14, rtol=0)
    xo.assert_allclose(p_test.px, p_ref.px, atol=1e-14, rtol=0)
    xo.assert_allclose(p_test.py, p_ref.py, atol=1e-14, rtol=0)
    xo.assert_allclose(p_test.delta, p_ref.delta, atol=1e-14, rtol=0)

    # Test backtracking
    line = xt.Line(elements=[mm])
    line.track(p_test, backtrack=True)
    xo.assert_allclose(p_test.s, 0.0, rtol=0, atol=1e-7)
    xo.assert_allclose(p_test.x, p0.x, atol=5e-14, rtol=0)
    xo.assert_allclose(p_test.y, p0.y, atol=5e-14, rtol=0)
    xo.assert_allclose(p_test.zeta, p0.zeta, atol=5e-14, rtol=0)
    xo.assert_allclose(p_test.px, p0.px, atol=5e-14, rtol=0)
    xo.assert_allclose(p_test.py, p0.py, atol=5e-14, rtol=0)
    xo.assert_allclose(p_test.delta, p0.delta, atol=1e-15, rtol=0)

# Check uniform integrator
mm1 = Magnet(h=0.1, k1=0.3, k0=0.2, length=2.0)
mm2 = mm1.copy()

mm1.edge_entry_active = False
mm1.edge_exit_active = False
mm2.edge_entry_active = False
mm2.edge_exit_active = False

mm1.integrator = 'uniform'
mm2.integrator = 'teapot'
mm1.num_multipole_kicks = 1
mm2.num_multipole_kicks = 1

p_test = p0.copy()
p_ref = p0.copy()

mm1.track(p_test)
mm2.track(p_ref)

xo.assert_allclose(p_test.s, 2.0, atol=0, rtol=1e-7)
xo.assert_allclose(p_ref.s, 2.0, atol=0, rtol=1e-7)
xo.assert_allclose(p_test.x, p_ref.x, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.y, p_ref.y, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.zeta, p_ref.zeta, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.px, p_ref.px, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.py, p_ref.py, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.delta, p_ref.delta, atol=1e-15, rtol=0)

# Test backtracking
line = xt.Line(elements=[mm1])
line.track(p_test, backtrack=True)
xo.assert_allclose(p_test.s, 0.0, atol=1e-7, rtol=0)
xo.assert_allclose(p_test.x, p0.x, atol=5e-14, rtol=0)
xo.assert_allclose(p_test.y, p0.y, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.zeta, p0.zeta, atol=1e-14, rtol=0)
xo.assert_allclose(p_test.px, p0.px, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.py, p0.py, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.delta, p0.delta, atol=1e-15, rtol=0)

# more kicks (needs loser thresholds)
mm1.num_multipole_kicks = 10
mm2.num_multipole_kicks = 10

p_test = p0.copy()
p_ref = p0.copy()

mm1.track(p_test)
mm2.track(p_ref)

xo.assert_allclose(p_test.s, 2.0, atol=0, rtol=1e-7)
xo.assert_allclose(p_ref.s, 2.0, atol=0, rtol=1e-7)
xo.assert_allclose(p_test.x, p_ref.x, atol=0, rtol=5e-3)
xo.assert_allclose(p_test.y, p_ref.y, atol=0, rtol=5e-3)
xo.assert_allclose(p_test.zeta, p_ref.zeta, atol=0, rtol=1e-2)
xo.assert_allclose(p_test.px, p_ref.px, atol=0, rtol=5e-3)
xo.assert_allclose(p_test.py, p_ref.py, atol=0, rtol=5e-3)
xo.assert_allclose(p_test.delta, p_ref.delta, atol=0, rtol=5e-3)

# Test backtracking
line = xt.Line(elements=[mm1])
line.track(p_test, backtrack=True)
xo.assert_allclose(p_test.s, 0.0, atol=1e-7, rtol=0)
xo.assert_allclose(p_test.x, p0.x, atol=5e-14, rtol=0)
xo.assert_allclose(p_test.y, p0.y, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.zeta, p0.zeta, atol=1e-14, rtol=0)
xo.assert_allclose(p_test.px, p0.px, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.py, p0.py, atol=1e-15, rtol=0)
xo.assert_allclose(p_test.delta, p0.delta, atol=1e-15, rtol=0)




# copyright ############################### #
# This file is part of the Xpart Package.   #
# Copyright (c) CERN, 2021.                 #
# ######################################### #
import pytest
import numpy as np
from scipy import constants

import xpart as xp
import xtrack as xt
from xobjects.test_helpers import fix_random_seed, for_all_test_contexts


def do_checks(test_context,part,n_part_per_bunch,sigma_z,sigma_delta,
        filled_slots,bunch_numbers,bunch_spacing):
        
    for i_bunch,bunch_number in enumerate(bunch_numbers):
        zeta_avg = np.average(
        test_context.nparray_from_context_array(
            part.zeta[i_bunch*n_part_per_bunch:
                   (i_bunch+1)*n_part_per_bunch]))
        delta_avg = np.average(
        test_context.nparray_from_context_array(
            part.delta[i_bunch*n_part_per_bunch:
                   (i_bunch+1)*n_part_per_bunch]))
        delta_rms = np.std(
            test_context.nparray_from_context_array(
                part.delta[i_bunch*n_part_per_bunch:
                           (i_bunch+1)*n_part_per_bunch]))
        zeta_rms = np.std(
            test_context.nparray_from_context_array(
                part.zeta[i_bunch*n_part_per_bunch:
                          (i_bunch+1)*n_part_per_bunch]))

        assert np.isclose((zeta_avg+bunch_spacing*filled_slots[bunch_number])/sigma_z, 0.0, atol=1e-2)
        assert np.isclose(delta_avg/sigma_delta, 0.0, atol=1e-2)
        assert np.isclose(zeta_rms, sigma_z, rtol=1e-2, atol=1e-15)
        assert np.isclose(delta_rms, sigma_delta, rtol=0.2, atol=1e-15)


@for_all_test_contexts
@fix_random_seed(64237673)
@pytest.mark.parametrize(
    'arc_type', ['arc_linear_fixed_qs', 'arc_linear_fixed_rf', 'arc_nonlinear']
)
def test_multi_bunch_gaussian_generation(test_context, arc_type):
    bunch_intensity = 1e11
    n_part_per_bunch = int(1e5)
    nemitt_x = 2e-6
    nemitt_y = 2.5e-6
    beta_x = 1.0
    beta_y = 1.0
    p0c = 7e12
    gamma = p0c/xp.PROTON_MASS_EV
    Q_x = 0.31
    Q_y = 0.32
    sigma_z = 0.08
    momentumCompaction = 3.483575072011584e-04
    eta = momentumCompaction - 1.0 / gamma ** 2
    circumference = 26658.883199999
    frev = constants.c/circumference
    f_RF = 400.8E6
    h = f_RF / frev
    betar = np.sqrt(1 - 1 / gamma ** 2)
    p0 = constants.m_p * betar * gamma * constants.c
    voltage = 16E6
    Qs = np.sqrt(constants.e * voltage * eta * h / (2 * np.pi * betar * constants.c * p0))
    beta_s = eta * circumference / (2 * np.pi * Qs);
    sigma_delta = sigma_z / beta_s
    bucket_length = circumference/h
    bunch_spacing_in_buckets = 10
    bunch_spacing = bunch_spacing_in_buckets * bucket_length
    filling_scheme = np.zeros(int(np.amin(h)/bunch_spacing_in_buckets))
    n_bunches_tot = 10
    filling_scheme[0:int(n_bunches_tot/2)] = 1
    filling_scheme[n_bunches_tot:int(3*n_bunches_tot/2)] = 1
    filled_slots = filling_scheme.nonzero()[0]
    n_procs = 2
    bunch_numbers_per_rank = xp.matched_gaussian.split_scheme(
        filling_scheme=filling_scheme,
        n_chunk=n_procs)

    if arc_type == 'arc_linear_fixed_qs':
        arc = xt.LineSegmentMap(
            betx=beta_x,qx=Q_x,
            bety=beta_y,qy=Q_y,
            bets=-beta_s,qs=Qs,bucket_length=bucket_length/constants.c,length=circumference)
    elif arc_type == 'arc_linear_fixed_rf':
        arc = xt.LineSegmentMap(_context=test_context,
            betx=beta_x,qx=Q_x,
            bety=beta_y,qy=Q_y,
            voltage_rf = voltage,
            longitudinal_mode = 'linear_fixed_rf',
            frequency_rf = f_RF,
            lag_rf = 180.0,
            slippage_length = circumference,
            momentum_compaction_factor = momentumCompaction,length=circumference)
    elif arc_type == 'arc_nonlinear':
        arc = xt.LineSegmentMap(_context=test_context,
            betx=beta_x,qx=Q_x,
            bety=beta_y,qy=Q_y,
            voltage_rf = voltage,
            longitudinal_mode = 'nonlinear',
            frequency_rf = f_RF,
            lag_rf = 180.0,
            slippage_length = circumference,
            momentum_compaction_factor = momentumCompaction,
            length=circumference)
    else:
        raise ValueError(f'Undefined test parameter for arc type {arc_type}')

    line = xt.Line([arc])
    line.particle_ref = xp.Particles(
        p0c=p0c,
        _context=test_context  # for testing purposes
    )
    line.build_tracker(_context=test_context)

    for rank in range(n_procs):
        part = xp.generate_matched_gaussian_multibunch_beam(
            _context=test_context,
            filling_scheme=filling_scheme,
            bunch_num_particles=n_part_per_bunch,
            bunch_intensity_particles=bunch_intensity,
            nemitt_x=nemitt_x, nemitt_y=nemitt_y, sigma_z=sigma_z,
            line=line, bunch_spacing_buckets=10,
            bunch_selection=bunch_numbers_per_rank[rank],
            particle_ref=line.particle_ref,
            bucket_length=bucket_length,
        )

        do_checks(test_context,part,n_part_per_bunch,sigma_z,sigma_delta,
            filled_slots,bunch_numbers_per_rank[rank],bunch_spacing)
        line.track(part,num_turns=100)
        do_checks(test_context,part,n_part_per_bunch,sigma_z,sigma_delta,
            filled_slots,bunch_numbers_per_rank[rank],bunch_spacing)



import xtrack as xt

env = xt.load_madx_lattice('../../test_data/pimms/PIMMS.seq')

line = env.pimms
line.particle_ref = xt.Particles(kinetic_energy0=100e6)

line['kqfa']=   0.30247
line['kqfb']=  0.523281
line['kqd']= -0.518932

line.twiss4d().plot()


import xtrack as xt

# Load a line and build a tracker
collider = xt.Environment.from_json(
    '../../test_data/hllhc15_thick/hllhc15_collider_thick.json')
collider.build_trackers()

tw0 = collider.twiss(method='4d')

twb1 = collider.lhcb1.twiss(start='e.ds.l5.b1', end='s.ds.r5.b1', init=tw0.lhcb1)
twb2 = collider.lhcb2.twiss(start='e.ds.l5.b2', end='s.ds.r5.b2', init=tw0.lhcb2)
vars = collider.vars
line_b1 = collider.lhcb1

opt = collider.match(
    solve=False,
    vary=xt.VaryList([
        'acbxv1.r5', 'acbxv1.l5', # <-- common elements
        'acbyvs4.l5b1', 'acbrdv4.r5b1', 'acbcv5.l5b1', 'acbcv6.r5b1', # <-- b1
        'acbyvs4.l5b2', 'acbrdv4.r5b2', 'acbcv5.r5b2', 'acbcv6.l5b2'  # <-- b2
        ],
        step=1e-10, limits=[-1e-3, 1e-3]),
    targets = [
        # Targets from b1 twiss
        twb1.target(y=0, py=10e-6, at='ip5'),
        twb1.target(y=0, py=0, at=xt.END),
        # Targets from b2 twiss
        twb2.target(y=0, py=-10e-6, at='ip5'),
        twb2.target(['y', 'py'], at=xt.END), # <-- preserve
        # Targets from vars
        vars.target('acbxv1.l5', xt.LessThan(1e-3)),
        vars.target('acbxv1.l5', xt.GreaterThan(1e-6)),
        vars.target(lambda vv: vv['acbxv1.l5'] + vv['acbxv1.r5'], xt.LessThan(1e-9)),
        # Targets from line
        line_b1.target(lambda ll: ll['mcbrdv.4r5.b1'].ksl[0], xt.GreaterThan(1e-6)),
        line_b1.target(lambda ll: ll['mcbxfbv.a2r5'].ksl[0] + ll['mcbxfbv.a2l5'].ksl[0],
                                xt.LessThan(1e-9)),
    ])
opt.solve()
opt.target_status()

# prints:
#
# Target status:
# id state tag tol_met      residue  current_val         target_val description
#  0 ON           True  6.95245e-17  6.95245e-17                  0 line=lhcb1, ('y', 'ip5'), val=0, tol=1e- ...
#  1 ON           True  -9.7917e-19        1e-05              1e-05 line=lhcb1, ('py', 'ip5'), val=1e-05, to ...
#  2 ON           True -3.83958e-16 -3.83958e-16                  0 line=lhcb1, ('y', 's.ds.r5.b1'), val=0,  ...
#  3 ON           True -1.73842e-17 -1.73842e-17                  0 line=lhcb1, ('py', 's.ds.r5.b1'), val=0, ...
#  4 ON           True -3.81775e-17 -3.81775e-17                  0 line=lhcb2, ('y', 'ip5'), val=0, tol=1e- ...
#  5 ON           True  -1.5128e-18       -1e-05             -1e-05 line=lhcb2, ('py', 'ip5'), val=-1e-05, t ...
#  6 ON           True -3.66729e-17 -3.66729e-17                0.0 line=lhcb2, ('y', 's.ds.r5.b2'), val=0,  ...
#  7 ON           True  1.35054e-18  1.35054e-18               -0.0 line=lhcb2, ('py', 's.ds.r5.b2'), val=-0 ...
#  8 ON           True            0        1e-06    LessThan(0.001) 'acbxv1.l5', val=LessThan(0.001), tol=1e ...
#  9 ON           True -8.47033e-22        1e-06 GreaterThan(1e-06) 'acbxv1.l5', val=GreaterThan(1e-06), tol ...
# 10 ON           True            0        1e-09    LessThan(1e-09) callable, val=LessThan(1e-09), tol=1e-10 ...
# 11 ON           True            0  1.02889e-06 GreaterThan(1e-06) line=lhcb1, callable, val=GreaterThan(1e ...
# 12 ON           True            0        1e-09    LessThan(1e-09) line=lhcb1, callable, val=LessThan(1e-09 ...

#!end-doc-part

import matplotlib.pyplot as plt

tw = collider.twiss()

plt.close('all')
fig = plt.figure(1, figsize=(6.4*1.3, 4.8))
plt.plot(tw.lhcb1.s - tw.lhcb1['s', 'ip5'], tw.lhcb1.y*1000, label='b1', color='blue')
plt.plot(tw.lhcb2.s - tw.lhcb2['s', 'ip5'], tw.lhcb2.y*1000, label='b2', color='red')
plt.xlabel('s [m]')
plt.ylabel('y [mm]')
plt.legend()

for nn in ['mcbxfbv.a2r5', 'mcbxfbv.a2l5', 'ip5',
           'mcbyv.a4l5.b1', 'mcbrdv.4r5.b1', 'mcbcv.5l5.b1',
           'mcbyv.4l5.b2', 'mcbrdv.4r5.b2', 'mcbcv.5r5.b2']:
    if nn.endswith('b2'):
        lname = 'lhcb2'
        ha = 'right'
        color = 'red'
    elif nn.endswith('b1'):
        lname = 'lhcb1'
        ha = 'left'
        color = 'blue'
    elif nn.startswith('mcbx'):
        lname = 'lhcb1'
        ha = 'right'
        color = 'green'
    else:
        lname = 'lhcb1'
        ha = 'left'
        color = 'black'
    plt.axvline(x=tw[lname]['s', nn] - tw[lname]['s', 'ip5'], color=color, linestyle='--', alpha=0.5)
    plt.text(tw[lname]['s', nn] - tw[lname]['s', 'ip5'], -1.1, nn, rotation=90,
            horizontalalignment=ha, verticalalignment='bottom', color=color)

plt.ylim(-1.15, 0.6)
plt.xlim(-250, 250)

import numpy as np
assert np.isclose(tw.lhcb1['y', 'ip5'], 0, rtol=0, atol=1e-9)
assert np.isclose(tw.lhcb1['py', 'ip5'], 10e-6, rtol=0, atol=1e-10)
assert np.isclose(tw.lhcb1['y', 's.ds.r5.b1'], 0, rtol=0, atol=1e-9)
assert np.isclose(tw.lhcb1['py', 's.ds.r5.b1'], 0, rtol=0, atol=1e-9)
assert np.isclose(tw.lhcb2['y', 'ip5'], 0, rtol=0, atol=1e-9)
assert np.isclose(tw.lhcb2['py', 'ip5'], -10e-6, rtol=0, atol=1e-10)
assert np.isclose(tw.lhcb2['y', 's.ds.r5.b2'], 0, rtol=0, atol=1e-9)
assert np.isclose(tw.lhcb2['py', 's.ds.r5.b2'], 0, rtol=0, atol=1e-9)

assert len(opt.targets) == 13
for ii in range(4):
    assert opt.targets[ii].action is opt.targets[0].action
    assert isinstance(opt.targets[ii].action, xt.match.ActionTwiss)
    assert opt.targets[ii].action.line.name == 'lhcb1'

for ii in range(4, 8):
    assert opt.targets[ii].action is opt.targets[4].action
    assert isinstance(opt.targets[ii].action, xt.match.ActionTwiss)
    assert opt.targets[ii].action.line.name == 'lhcb2'

assert isinstance(opt.targets[8].action, xt.line.ActionVars)
assert isinstance(opt.targets[9].action, xt.line.ActionVars)
assert isinstance(opt.targets[10].action, xt.line.ActionVars)
assert isinstance(opt.targets[11].action, xt.line.ActionLine)
assert isinstance(opt.targets[12].action, xt.line.ActionLine)

opt.tag('solution')
opt.reload(0)

assert np.all(opt.target_status(ret=True).tol_met == np.array(
    [ True, False,  True,  True,  True, False,  True,  True,  True,
      False,  True, False,  True]))
opt.reload(tag='solution')
assert np.all(opt.target_status(ret=True).tol_met)
plt.show()



import xtrack as xt
import xobjects as xo
import numpy as np

env = xt.Environment()

line0 = env.new_line(
    components=[
        env.new('q0', 'Quadrupole', length=2.0, at=20.0),
        env.new('ql', 'Quadrupole', length=2.0, at=-10.0, from_='q0'),
        env.new('qr', 'Quadrupole', length=2.0, at=10.0, from_='q0'),
        env.new('mk1', 'Marker', at=40),
        env.new('mk2', 'Marker'),
        env.new('mk3', 'Marker'),
        env.place('q0'),
        env.new('end', 'Marker', at=50.),
    ])
tt0 = line0.get_table()
tt0.show(cols=['name', 's_start', 's_end', 's_center'])

line1 = line0.copy()
line1.remove('q0::1')
tt1 = line1.get_table()
tt1.show(cols=['name', 's_start', 's_end', 's_center'])

assert np.all(tt1.name == np.array(
    ['drift_1', 'ql', 'drift_2', 'q0', 'drift_3', 'qr', 'drift_4',
       'mk1', 'mk2', 'mk3', 'drift_6', 'drift_5', 'end', '_end_point']))
xo.assert_allclose(tt1.s_center, np.array(
    [ 4.5, 10. , 15. , 20. , 25. , 30. , 35.5, 40. , 40. , 40. , 41. ,
       46. , 50. , 50. ]), rtol=0., atol=1e-14)

line2 = line0.copy()
line2.remove('q0')
tt2 = line2.get_table()
tt2.show(cols=['name', 's_start', 's_end', 's_center'])

assert np.all(tt2.name == np.array(
    ['drift_1', 'ql', 'drift_2', 'drift_6', 'drift_3', 'qr', 'drift_4',
       'mk1', 'mk2', 'mk3', 'drift_7', 'drift_5', 'end', '_end_point']))
xo.assert_allclose(tt2.s_center, np.array(
    [ 4.5, 10. , 15. , 20. , 25. , 30. , 35.5, 40. , 40. , 40. , 41. ,
       46. , 50. , 50. ]), rtol=0., atol=1e-14)

line3 = line0.copy()
line3.remove('q.*')
tt3 = line3.get_table()
tt3.show(cols=['name', 's_start', 's_end', 's_center'])

assert np.all(tt3.name == np.array(
    ['drift_1', 'drift_6', 'drift_2', 'drift_7', 'drift_3', 'drift_8',
     'drift_4', 'mk1', 'mk2', 'mk3', 'drift_9', 'drift_5', 'end',
     '_end_point']))
xo.assert_allclose(tt3.s_center, np.array(
    [ 4.5, 10. , 15. , 20. , 25. , 30. , 35.5, 40. , 40. , 40. , 41. ,
       46. , 50. , 50. ]), rtol=0., atol=1e-14)

line4 = line0.copy()
line4.remove('mk2')
tt4 = line4.get_table()
tt4.show(cols=['name', 's_start', 's_end', 's_center'])

assert np.all(tt4.name == np.array(
    ['drift_1', 'ql', 'drift_2', 'q0::0', 'drift_3', 'qr', 'drift_4',
       'mk1', 'mk3', 'q0::1', 'drift_5', 'end', '_end_point']))
xo.assert_allclose(tt4.s_center, np.array(
    [ 4.5, 10. , 15. , 20. , 25. , 30. , 35.5, 40. , 40. , 41. , 46. ,
       50. , 50. ]), rtol=0., atol=1e-14)

line5 = line0.copy()
line5.remove('mk.*')
tt5 = line5.get_table()
tt5.show(cols=['name', 's_start', 's_end', 's_center'])

assert np.all(tt5.name == np.array(
    ['drift_1', 'ql', 'drift_2', 'q0::0', 'drift_3', 'qr', 'drift_4',
       'q0::1', 'drift_5', 'end', '_end_point']))
xo.assert_allclose(tt5.s_center, np.array(
    [ 4.5, 10. , 15. , 20. , 25. , 30. , 35.5, 41. , 46. , 50. , 50. ]),
    rtol=0., atol=1e-14)


import numpy as np
from cpymad.madx import Madx
import xtrack as xt
import os

# TODO:
# - handle thick slicing of bends (edges!)
# - handle isolated dipole edges
# - crab cavities
# - rbarc (sps)

# ----- Test sequence -----
# mad = Madx()
# # Element definitions
# mad.input("""

# a = 1.;
# b := sin(3*a) + cos(2*a);

# cav1: rfcavity, freq:=a*10, lag:=a*0.5, volt:=a*6;
# cav2: rfcavity, freq:=10, lag:=0.5, volt:=6;
# testseq: sequence, l=10;
# c1: cav1, at=0.2, apertype=circle, aperture=0.01;
# c2: cav2, at=0.5, apertype=circle, aperture=0.01;
# endsequence;
# """
# )
# # Beam
# mad.input("""
# beam, particle=proton, gamma=1.05, sequence=testseq;
# """)
# mad.use('testseq')
# seq = mad.sequence['testseq']
# line = xt.Line.from_madx_sequence(sequence=seq, deferred_expressions=True)

# ----- Elena -----
# mad = Madx()
# folder = ('../../test_data/elena')
# mad.call(folder + '/elena.seq')
# mad.call(folder + '/highenergy.str')
# mad.call(folder + '/highenergy.beam')
# mad.use('elena')
# seq = mad.sequence.elena
# line = xt.Line.from_madx_sequence(seq)
# line.particle_ref = xt.Particles(gamma0=seq.beam.gamma,
#                                     mass0=seq.beam.mass * 1e9,
#                                     q0=seq.beam.charge)

# line.slice_thick_elements(
#     slicing_strategies=[
#         xt.Strategy(slicing=None), # don't touch other elements
#         xt.Strategy(slicing=xt.Uniform(10, mode='thick'), element_type=xt.Bend)
#     ])

# ----- LHC (thick) -----
line = xt.Line.from_json('../../test_data/hllhc15_thick/lhc_thick_with_knobs.json')
line.build_tracker()

# ----- LHC (thin) -----
# line = xt.Line.from_json('../../test_data/hllhc15_noerrors_nobb/line_w_knobs_and_particle.json')
# line.particle_ref = xt.Particles(mass=xt.PROTON_MASS_EV, p0c=7000e9)
# line.build_tracker()

# ----- LHC (thin) -----
# mad1 = Madx()
# mad1.call('../../test_data/hllhc15_noerrors_nobb/sequence_with_crabs.madx')
# mad1.use('lhcb1')
# seq = mad1.sequence.lhcb1
# line = xt.Line.from_madx_sequence(seq, deferred_expressions=True)
# line.particle_ref = xt.Particles(gamma0=seq.beam.gamma,
#                                     mass0=seq.beam.mass * 1e9,
#                                     q0=seq.beam.charge)

mad_seq = line.to_madx_sequence(sequence_name='myseq')

mad2 = Madx()
mad2.input(mad_seq)
mad2.beam(particle='proton', energy=7000e9) #!!!!!
mad2.use('myseq')

line2 = xt.Line.from_madx_sequence(mad2.sequence.myseq, deferred_expressions=True)

temp_fname = 'temp4madng'
with open(temp_fname+'.madx', 'w') as fid:
    fid.write(mad_seq)

# # Use open twiss to compare
# tm1 = xt.Table(mad1.twiss(betx=1, bety=1,x=1e-3, y=1e-3))
# tm2 = xt.Table(mad2.twiss(betx=1, bety=1,x=1e-3, y=1e-3))

# # tx1 = line.twiss(method='4d')
# # tx2 = line2.twiss(method='4d')

# tm1 = xt.Table(mad1.twiss())
# tm2 = xt.Table(mad2.twiss())

# mux2_interp = np.interp(tm1.s, tm2.s, tm2.mux)
# dx2_interp = np.interp(tm1.s, tm2.s, tm2.dx)



import xtrack as xt

env = xt.Environment()
env.vars.default_to_zero = True
line = env.new_line(components=[
    env.new('mq', 'Quadrupole', length=0.5, k1='kq'),
    env.new('mqs', 'Quadrupole', length=0.5, k1s='kqs'),
    env.new('mb', 'Bend', length=0.5, angle='ang', k0_from_h=True),
])

env.set_multipolar_errors({
    'mq': {'rel_knl': [1e-6, 1e-5, 1e-4], 'rel_ksl': [-1e-6, -1e-5, -1e-4]},
    'mqs': {'rel_knl': [2e-6, 2e-5, 2e-4], 'rel_ksl': [3e-6, 3e-5, 3e-4], 'refer': 'k1s'},
    'mb': {'rel_knl': [2e-6, 3e-5, 4e-4], 'rel_ksl': [5e-6, 6e-5, 7e-4]},
})

# Errors respond when variables are changed
env['kq'] = 0.1
env['kqs'] = 0.2
env['ang'] = 0.3

# inspect
env.ref['mq'].knl[1]._info()
# is:
# #  element_refs['mq'].knl[1]._get_value()
#    element_refs['mq'].knl[1] = 5.000000000000001e-07

# #  element_refs['mq'].knl[1]._expr
#    element_refs['mq'].knl[1] = (0.0 + ((vars['err_mq_knl1'] *
#                          element_refs['mq'].k1) * element_refs['mq'].length))

# #  element_refs['mq'].knl[1]._expr._get_dependencies()
#    element_refs['mq'] = Quadrupole(k1=0.1, k1s=0, ...
#    vars['err_mq_knl1'] = 1e-05
#    element_refs['mq'].k1 = 0.1
#    element_refs['mq'].length = 0.5

#  element_refs['mq'].knl[1] does not influence any target

import xobjects as xo
import numpy as np

xo.assert_allclose(env.get('mq').knl[:3], 0.5 * 0.1 * np.array([1e-6, 1e-5, 1e-4]), rtol=1e-7, atol=0)
xo.assert_allclose(env.get('mq').ksl[:3], 0.5 * 0.1 * np.array([-1e-6, -1e-5, -1e-4]), rtol=1e-7, atol=0)
xo.assert_allclose(env.get('mqs').knl[:3], 0.5 * 0.2 * np.array([2e-6, 2e-5, 2e-4]), rtol=1e-7, atol=0)
xo.assert_allclose(env.get('mqs').ksl[:3], 0.5 * 0.2 * np.array([3e-6, 3e-5, 3e-4]), rtol=1e-7, atol=0)
xo.assert_allclose(env.get('mb').knl[:3], 0.3 * np.array([2e-6, 3e-5, 4e-4]), rtol=1e-7, atol=0)
xo.assert_allclose(env.get('mb').ksl[:3], 0.3 * np.array([5e-6, 6e-5, 7e-4]), rtol=1e-7, atol=0)
